import numpy as np

from lingam.tools import (
    bootstrap_with_imputation,
    BaseMultipleImputation,
    BaseMultiGroupCDModel,
)


def test_bootstrap_with_imputation():
    # test data
    m = np.array([
        [ 0.000,  0.000,  0.000,  0.895,  0.000,  0.000],
        [ 0.565,  0.000,  0.377,  0.000,  0.000,  0.000],
        [ 0.000,  0.000,  0.000,  0.895,  0.000,  0.000],
        [ 0.000,  0.000,  0.000,  0.000,  0.000,  0.000],
        [ 0.991,  0.000, -0.124,  0.000,  0.000,  0.000],
        [ 0.895,  0.000,  0.000,  0.000,  0.000,  0.000]
    ])
    error_vars = [0.2, 0.2, 0.2, 1.0, 0.2, 0.2]
    n_samples = 1000
    X = _generate_data(m, error_vars, sample_size=n_samples)

    # test data with missing values by MCAR
    prop_list = [0, 0, 0, 0, 0, 0.05]
    X_mcar = _add_nan_mcar(X, prop_list=prop_list)
    
    # default
    n_sampling = 5
    n_repeats = 3
    causal_orders, adj_matrices_list, resampled_indices, imputation_results = bootstrap_with_imputation(
        X_mcar, n_sampling, n_repeats, random_state=0)

    # check shapes
    assert causal_orders.shape == (n_sampling, len(m))
    assert adj_matrices_list.shape == (n_sampling, n_repeats, *m.shape)
    assert resampled_indices.shape == (n_sampling, n_samples)
    assert imputation_results.shape == (n_sampling, n_repeats, *X_mcar.shape)

    # check causal orders
    assert np.all([np.unique(r) == np.arange(len(m)) for r in causal_orders])

    # confirm imputed positions
    for i, imputation_result in enumerate(imputation_results):
        indices = resampled_indices[i]
        resampled = X_mcar[indices]
        nan_pos = np.argwhere(np.isnan(resampled))

        for result in imputation_result:
            imputation_pos = np.argwhere(~np.isnan(result))
            assert (nan_pos == imputation_pos).all()


def test_bootstrap_with_imputation_imp():
    """ tests for the function to check the output of the given imputer. """

    # test data
    m = np.array([
        [0, 0, 0],
        [1, 0, 0],
        [0, 1, 0],
    ]).astype(float)
    error_vars = np.ones(3)
    n_samples = 10
    X_org = _generate_data(m, error_vars, sample_size=n_samples)
    X_mcar = _add_nan_mcar(X_org, prop_list=None)

    n_features = len(m)
    n_sampling = 1
    n_repeats = 2

    # valid
    class ValidImputer(BaseMultipleImputation):
        def fit_transform(self, X):
            imputeds = np.array([X_org for _ in range(n_repeats)])
            return imputeds

    imp = ValidImputer()
    result = bootstrap_with_imputation(X_mcar, n_sampling, imp=imp)

    causal_order, adj_matrices_list, resampled_indices, imputation_results = result
    assert causal_order.shape == (n_sampling, len(m))
    assert adj_matrices_list.shape == (n_sampling, n_repeats, *m.shape)
    assert resampled_indices.shape == (n_sampling, n_samples)
    assert imputation_results.shape == (n_sampling, n_repeats, *X_mcar.shape)

    # invalid: not inherited
    class InvalidImputer:
        def fit_transform(self, X):
            imputeds = np.array([X_org for _ in range(n_repeats)])
            return imputeds

    imp = InvalidImputer()
    try:
        bootstrap_with_imputation(X_mcar, n_sampling, imp=imp)
    except Exception as e:
        pass
    else:
        raise AssertionError

    # invalid: exists nan 
    class InvalidImputer(BaseMultipleImputation):
        def fit_transform(self, X):
            imputeds = np.array([X_mcar for _ in range(n_repeats)])
            return imputeds

    imp = InvalidImputer()
    try:
        bootstrap_with_imputation(X_mcar, n_sampling, imp=imp)
    except Exception as e:
        pass
    else:
        raise AssertionError

    # invalid: wrong shape
    class InvalidImputer(BaseMultipleImputation):
        def fit_transform(self, X):
            imputeds = np.array([X_org for _ in range(n_repeats)])
            imputeds = imputeds[:, :-1, :]
            return imputeds

    imp = InvalidImputer()
    try:
        bootstrap_with_imputation(X_mcar, n_sampling, imp=imp)
    except Exception as e:
        pass
    else:
        raise AssertionError

    # invalid: wrong data type
    class InvalidImputer(BaseMultipleImputation):
        def fit_transform(self, X):
            imputeds = np.array([X_org for _ in range(n_repeats)])
            return imputeds.astype(str)

    imp = InvalidImputer()
    try:
        bootstrap_with_imputation(X_mcar, n_sampling, imp=imp)
    except Exception as e:
        pass
    else:
        raise AssertionError

def test_bootstrap_with_imputation_cd_model():
    """ tests for the function to check the output of the given cd_model """

    # test data
    m = np.array([
        [0, 0, 0],
        [1, 0, 0],
        [0, 1, 0],
    ]).astype(float)
    error_vars = np.ones(3)
    n_samples = 10
    X = _generate_data(m, error_vars, sample_size=n_samples)
    X_mcar = _add_nan_mcar(X, prop_list=None)

    n_features = len(m)
    n_sampling = 1
    n_repeats = 2

    # valid
    class ValidCDModel(BaseMultiGroupCDModel):
        def before_imputation(self, X):
            pass
        def fit(self, X_list):
            co = [0, 1, 2]
            ams = [m for _ in range(len(X_list))]
            return co, ams

    n_sampling = 2
    cd_model = ValidCDModel()
    result = bootstrap_with_imputation(X_mcar, n_sampling, n_repeats=n_repeats, cd_model=cd_model)

    causal_order, adj_matrices_list, resampled_indices, imputation_results = result
    assert causal_order.shape == (n_sampling, len(m))
    assert adj_matrices_list.shape == (n_sampling, n_repeats, *m.shape)
    assert resampled_indices.shape == (n_sampling, n_samples)
    assert imputation_results.shape == (n_sampling, n_repeats, *X_mcar.shape)

    # valid: cd_model estimates one adj_matrix based on X_list
    class ValidCDModel(BaseMultiGroupCDModel):
        def before_imputation(self, X):
            pass
        def fit(self, X_list):
            co = [0, 1, 2]
            ams = m.copy()
            return co, ams

    n_sampling = 2
    cd_model = ValidCDModel()
    result = bootstrap_with_imputation(X_mcar, n_sampling, n_repeats=n_repeats, cd_model=cd_model)

    causal_order, adj_matrices_list, resampled_indices, imputation_results = result
    assert causal_order.shape == (n_sampling, len(m))
    assert adj_matrices_list.shape == (n_sampling, *m.shape)
    assert resampled_indices.shape == (n_sampling, n_samples)
    assert imputation_results.shape == (n_sampling, n_repeats, *X_mcar.shape)

    # valid: adj_matrices: contains nan
    class ValidCDModel(BaseMultiGroupCDModel):
        def before_imputation(self, X):
            pass
        def fit(self, X_list):
            co = [0, 1, 2]
            ams = [m.copy() for _ in range(len(X_list))]
            ams[0][0, 0] = np.nan
            return (co, ams)

    n_sampling = 2
    cd_model = ValidCDModel()
    result = bootstrap_with_imputation(X_mcar, n_sampling, n_repeats=n_repeats, cd_model=cd_model)
    
    causal_order, adj_matrices_list, resampled_indices, imputation_results = result
    assert causal_order.shape == (n_sampling, len(m))
    assert adj_matrices_list.shape == (n_sampling, n_repeats, *m.shape)
    assert resampled_indices.shape == (n_sampling, n_samples)
    assert imputation_results.shape == (n_sampling, n_repeats, *X_mcar.shape)

    # invalid: not inherited
    class InvalidCDModel:
        def before_imputation(self, X):
            pass
        def fit(self, X_list):
            co = [0, 1, 2]
            ams = [m for _ in range(len(X_list))]
            return co, ams

    cd_model = InvalidCDModel()
    try:
        bootstrap_with_imputation(X_mcar, n_sampling, n_repeats=n_repeats, cd_model=cd_model)
    except Exception as e:
        pass
    else:
        raise AssertionError

    # invalid: wrong type
    class InvalidCDModel(BaseMultiGroupCDModel):
        def before_imputation(self, X):
            pass
        def fit(self, X_list):
            co = [0, 1, 2]
            ams = [m for _ in range(len(X_list))]
            return [co, ams]

    cd_model = InvalidCDModel()
    try:
        bootstrap_with_imputation(X_mcar, n_sampling, n_repeats=n_repeats, cd_model=cd_model)
    except Exception as e:
        pass
    else:
        raise AssertionError

    # invalid: wrong length
    class InvalidCDModel(BaseMultiGroupCDModel):
        def before_imputation(self, X):
            pass
        def fit(self, X_list):
            co = [0, 1, 2]
            ams = [m for _ in range(len(X_list))]
            return (co, ams, 0)

    cd_model = InvalidCDModel()
    try:
        bootstrap_with_imputation(X_mcar, n_sampling, n_repeats=n_repeats, cd_model=cd_model)
    except Exception as e:
        pass
    else:
        raise AssertionError

    # invalid: causal_order: contains nan
    class InvalidCDModel(BaseMultiGroupCDModel):
        def before_imputation(self, X):
            pass
        def fit(self, X_list):
            co = [np.nan, 1, 2]
            ams = [m for _ in range(len(X_list))]
            return (co, ams)

    cd_model = InvalidCDModel()
    try:
        bootstrap_with_imputation(X_mcar, n_sampling, n_repeats=n_repeats, cd_model=cd_model)
    except Exception as e:
        pass
    else:
        raise AssertionError

    # invalid: causal_order: wrong length
    class InvalidCDModel(BaseMultiGroupCDModel):
        def before_imputation(self, X):
            pass
        def fit(self, X_list):
            co = [0, 1, 2, 3]
            ams = [m for _ in range(len(X_list))]
            return (co, ams)

    cd_model = InvalidCDModel()
    try:
        bootstrap_with_imputation(X_mcar, n_sampling, n_repeats=n_repeats, cd_model=cd_model)
    except Exception as e:
        pass
    else:
        raise AssertionError

    # invalid: causal_order: not unique
    class InvalidCDModel(BaseMultiGroupCDModel):
        def before_imputation(self, X):
            pass
        def fit(self, X_list):
            co = [0, 0, 1]
            ams = [m for _ in range(len(X_list))]
            return (co, ams)

    cd_model = InvalidCDModel()
    try:
        bootstrap_with_imputation(X_mcar, n_sampling, n_repeats=n_repeats, cd_model=cd_model)
    except ValueError as e:
        pass
    else:
        raise AssertionError

    # invalid: adj_matrices: wrong type
    class InvalidCDModel(BaseMultiGroupCDModel):
        def before_imputation(self, X):
            pass
        def fit(self, X_list):
            co = [0, 1, 2]
            ams = [m.copy().astype(str) for _ in range(len(X_list))]
            return (co, ams)

    cd_model = InvalidCDModel()
    try:
        bootstrap_with_imputation(X_mcar, n_sampling, n_repeats=n_repeats, cd_model=cd_model)
    except Exception as e:
        pass
    else:
        raise AssertionError

    # invalid: adj_matrices: wrong shape
    class InvalidCDModel(BaseMultiGroupCDModel):
        def before_imputation(self, X):
            pass
        def fit(self, X_list):
            co = [0, 1, 2]
            ams = [m.copy()[:-1, :] for _ in range(len(X_list))]
            return (co, ams)

    cd_model = InvalidCDModel()
    try:
        bootstrap_with_imputation(X_mcar, n_sampling, n_repeats=n_repeats, cd_model=cd_model)
    except Exception as e:
        pass
    else:
        raise AssertionError

def _generate_data(m, error_vars, sample_size=1000):
    params = [0.5 * np.sqrt(12 * v) for v in error_vars]
    generate_error = lambda p: np.random.uniform(-p, p, size=sample_size)
    e = np.array([generate_error(p) for p in params])
    
    X = np.linalg.pinv(np.eye(len(m)) - m) @ e
    X = X.T

    return X

def _add_nan_mcar(X, prop_list=None):
    X_mcar = X.copy()

    if prop_list is None:
        prop_list = np.ones(X_mcar.shape[1]) * 0.1

    for i, prop in enumerate(prop_list):
        mask = np.random.uniform(0, 1, size=len(X_mcar))
        X_mcar[mask < prop, i] = np.nan

    return X_mcar





import pytest

import numpy as np
import pandas as pd

from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.preprocessing import OneHotEncoder
from sklearn.compose import ColumnTransformer
from sklearn.pipeline import Pipeline
from sklearn.metrics import mean_absolute_error
from scipy.special import expit

import os, sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from lingam.experimental import CausalDataGenerator

@pytest.fixture
def init():
    return lambda :np.random.seed(0)

@pytest.fixture
def test_data():
    np.random.seed(0)

    N=10000

    errors = pd.DataFrame({
        "x1": np.random.uniform(size=N),
        "x2": np.random.uniform(size=N),
        "x3": np.random.uniform(size=N),
        "x4": np.random.uniform(size=N),
    })
    
    def model_x1(df, error, args):
        return 2 * df["x3"] - 2 * df["x4"] + error
    
    def model_x2(df, error, args):
        return 2 * df["x1"] - 3 * df["x3"] - 3 * df["x4"] + error
        
    def model_x3(df, error, args):
        return error
    
    def model_x4(df, error, args):
        return error
    
    models = {
        "x1": model_x1,
        "x2": model_x2,
        "x3": model_x3,
        "x4": model_x4,
    }
    
    causal_order = ["x3", "x4", "x1", "x2"]
    
    X = pd.DataFrame(np.nan, index=np.arange(N), columns=list(errors.keys()))
    for name in causal_order:
        X[name] = models[name](X, errors[name], None)

    return X, models, errors, causal_order

@pytest.fixture
def test_data2():
    np.random.seed(0)

    N=10000

    errors = pd.DataFrame({
        "x1": np.random.uniform(size=N),
        "x2": np.random.uniform(size=N),
        "x3": np.random.choice(["a", "b"], size=N),
        "x4": np.random.uniform(size=N),
    })
    
    def model_x1(df, error, args):
        def func(s):
            return 1.5 * s["x4"] + (-2 if s["x3"] == "a" else 0.5) + errors.loc[s.name, "x1"]
        
        p = df[["x3", "x4"]].apply(func, axis=1)
        d = expit(p).apply(lambda v: "c" if v > 0.5 else "d")
        return pd.Categorical(d)
            
    def model_x2(df, error, args):
        w_map = {
            ("a", "c"): 2,
            ("a", "d"): -2,
            ("b", "c"): 0.5,
            ("b", "d"): -0.5,
        }
        def func(s):
            return w_map[s["x3"], s["x1"]] * s["x4"]
        
        d = df[["x1", "x3", "x4"]].apply(func, axis=1) + error
        return d
        
    def model_x3(df, error, args):
        return pd.Categorical(errors["x3"])
    
    def model_x4(df, error, args):
        return errors["x4"]
    
    models = {
        "x1": model_x1,
        "x2": model_x2,
        "x3": model_x3,
        "x4": model_x4,
    }
    
    causal_order = ["x3", "x4", "x1", "x2"]
    
    X = pd.DataFrame(np.nan, index=np.arange(N), columns=list(errors.keys()))
    for name in causal_order:
        X[name] = models[name](X, errors[name], None)
    
    return X, models, errors, causal_order

def test_continuous(init, test_data):
    init()

    X, models, errors, causal_order = test_data

    g = CausalDataGenerator()
    g.fit(X, "x1", "x2", adjustments=["x3", "x4"], model_cause=LinearRegression(), model_effect=LinearRegression())

    # generate data using ground truth functions and noises.
    generated = g.generate(interv_exog=errors.to_dict(orient="list"), cause_model=models["x1"], effect_model=models["x2"])
    mae = mean_absolute_error(X, generated)
    if not np.isclose(mae, 0):
        raise ValueError

def test_discrete(init, test_data2):
    init()

    X, models, errors, causal_order = test_data2

    # models
    transformers = [
        ("categorical", OneHotEncoder(sparse_output=False), ["x3"]),
        ("numeric", "passthrough", ["x4"]),
    ]
    trans = ColumnTransformer(transformers=transformers)
    
    model_cause = Pipeline([
        ("transformer", trans),
        ("estimator", LogisticRegression())
    ])
    
    transformers = [
        ("categorical", OneHotEncoder(sparse_output=False), ["x1", "x3"]),
        ("numeric", "passthrough", ["x4"]),
    ]
    trans = ColumnTransformer(transformers=transformers)
    
    model_effect = Pipeline([
        ("transformer", trans),
        ("estimator", LinearRegression())
    ])

    g = CausalDataGenerator()
    g.fit(X, "x1", "x2", adjustments=["x3", "x4"], model_cause=model_cause, model_effect=model_effect)

    # generate data using ground truth functions and noises.
    generated = g.generate(interv_exog=errors.to_dict(orient="list"), cause_model=models["x1"], effect_model=models["x2"])
    mae = mean_absolute_error(X[["x2", "x4"]], generated[["x2", "x4"]])
    if not np.isclose(mae, 0):
        raise ValueError

    s = (X[["x1", "x3"]] == generated[["x1", "x3"]]).sum().sum()
    if s != X.shape[0] * 2:
        raise ValueError



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""

import itertools
import warnings
import numpy as np
from sklearn.utils import check_array, check_scalar
from sklearn.linear_model import LassoLarsCV, LinearRegression
from sklearn.preprocessing import StandardScaler

from .direct_lingam import DirectLiNGAM


class HighDimDirectLiNGAM(DirectLiNGAM):
    """ An implementation of the high-dimensional LiNGAM algorithm. [1]_

    References
    ----------
    .. [1] Wang, Y. Samuel, and Mathias Drton. "High-dimensional causal discovery under non-Gaussianity."
       Biometrika 107.1 (2020): 41-59.
    """

    def __init__(self, J=3, K=4, alpha=0.5, estimate_adj_mat=True, random_state=None):
        """Construct a HighDimDirectLiNGAM model.

        Parameters
        ----------
        J : int, optional (default=3)
            Assumed largest in-degree.
        K : int, optional (default=4)
            The degree of the moment which is non-Gaussianity.
        alpha : float, optional (default=0.5)
            The value for pruning away false parents.
        estimate_adj_mat : bool, optional (default=True)
            If Fase, skip the estimation of the adjacency matrix.
        random_state : int, optional (default=None)
            ``random_state`` is the seed used by the random number generator.
        """

        super().__init__(random_state=random_state)

        self._J = check_scalar(J, "J", int, min_val=2, include_boundaries="neither")
        self._K = check_scalar(K, "K", int, min_val=1)
        self._alpha = check_scalar(alpha, "alpha", (int, float), min_val=0, max_val=1)
        self._estimate_adj_mat = check_scalar(estimate_adj_mat, "estimate_adj_mat", bool)

    def fit(self, X):
        """Fit the model to X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.

        Returns
        -------
        self : object
            Fitted model.
        """

        self._Y = check_array(X)

        self._yty = self._Y.T @ self._Y

        cut_off = 0

        theta = []
        p = self._Y.shape[1]

        psi = list(range(p))

        prune_stats = np.full((p, p), 1e5)
        np.fill_diagonal(prune_stats, 0)

        while len(psi) > 1:
            new_stats = []
            for v in psi:
                cond_set = set.intersection(set(theta), set(np.argwhere(prune_stats[v] > cut_off).flatten()))
                cond_set = set.union(cond_set, set(theta[-1:] if len(theta) > 0 else set()))

                last_root = theta[-1] if len(theta) > 0 else None
                stats = self._get_prune_stats(v, psi, self._K, last_root, cond_set, self._J)

                new_stats.append(stats)
            new_stats = np.array(new_stats)

            prune_stats[psi, :] = np.min([prune_stats[psi, :], new_stats], axis=0)
            np.fill_diagonal(prune_stats, 0)

            max_taus = np.max(prune_stats[np.ix_(psi, psi)], axis=1)

            r = psi[np.argmin(max_taus)]

            cut_off = max(cut_off, min(max_taus) * self._alpha)

            theta.append(r)
            psi.remove(r)

        self._causal_order = [*theta, *psi]

        if not self._estimate_adj_mat:
            return self

        if self._Y.shape[0] <= self._Y.shape[1]:
            warnings.warn(
                "Since n_samples <= n_features, the adjacency matrix is "
                "estimated with LassoLarsCV instead of LassoLarsIC."
            )
            return self._estimate_adjacency_matrix2(self._Y)

        return self._estimate_adjacency_matrix(self._Y)

    def _calc_tau(self, k, pa, ch):
        return abs(np.mean((pa ** (k - 1)) * ch) * np.mean(pa ** 2) - np.mean(pa ** k) * np.mean(pa * ch))

    def _calc_taus(self, pa, ch, k, cond_sets, an_sets):
        n = self._Y.shape[0]
        ret = np.empty(self._Y.shape[1])
        ret.fill(1e10)

        for z in range(len(cond_sets)):
            cond = cond_sets[z]

            b = np.linalg.pinv(self._yty[np.ix_(cond, cond)]) @ self._yty[cond, pa]
            resid = (self._Y[:, [pa]] - self._Y[:, cond] @ b.reshape(len(cond), 1)).flatten()

            resid_k_1 = np.power(resid, k - 1)
            resid_var = np.mean(np.power(resid, 2))
            resid_k = np.mean(np.power(resid, k))

            for i in range(len(ch)):
                value = (1.0 / n) * np.dot(resid_k_1, self._Y[:, ch[i]]) * resid_var\
                    - resid_k * (1.0 / n) * np.dot(resid, self._Y[:, ch[i]])
                ret[ch[i]] = np.min([ret[ch[i]], abs(value)])

            for i in range(an_sets.shape[1]):
                value = (1.0 / n) * np.dot(resid_k_1.T, self._Y[:, an_sets[z, i]]) * resid_var\
                    - resid_k * (1.0 / n) * np.dot(resid.T, self._Y[:, an_sets[z, i]])
                ret[an_sets[z, i]] = np.min([ret[an_sets[z, i]], abs(value)])

            return ret

    def _get_prune_stats(self, i, j, K, last_root, condition_set, J):
        j = list(set(j) - {i})
        p = self._Y.shape[1]
        prune_stat = np.array([1e5 for _ in range(p)])

        if last_root is None:
            prune_stat[j] = [self._calc_tau(K, self._Y[:, i], self._Y[:, j_]) for j_ in j]
            return prune_stat

        size_of_set = min(J, len(condition_set))

        if len(condition_set - {last_root}) == 1:
            condition_sub_set = np.array([list(condition_set - {last_root})]).reshape(1, 1)
        else:
            condition_sub_set = list(itertools.combinations(list(condition_set - {last_root}), size_of_set - 1))
            condition_sub_set = np.array([list(e) for e in condition_sub_set])

        an_sets = np.array([list(condition_set - set(x)) for x in condition_sub_set])
        if len(an_sets) == 1 and an_sets.shape[1] > 0:
            an_sets = an_sets.T

        condition_sub_set = np.concatenate([
            np.full(len(condition_sub_set), last_root).reshape(-1, 1),
            condition_sub_set
        ], axis=1).astype(int)

        prune_stat = self._calc_taus(i, j, K, condition_sub_set, an_sets)
        return prune_stat

    def _estimate_adjacency_matrix2(self, X):
        """Estimate adjacency matrix by causal order.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.

        Returns
        -------
        self : object
            Returns the instance itself.
        """

        B = np.zeros([X.shape[1], X.shape[1]], dtype="float64")
        for i in range(1, len(self._causal_order)):
            target = self._causal_order[i]
            predictors = self._causal_order[:i]

            # target is exogenous variables if predictors are empty
            if len(predictors) == 0:
                continue

            B[target, predictors] = self._predict_adaptive_lasso(X, predictors, target)

        self._adjacency_matrix = B
        return self

    def _predict_adaptive_lasso(self, X, predictors, target, gamma=1.0):
        """Predict with Adaptive Lasso.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.
        predictors : array-like, shape (n_predictors)
            Indices of predictor variable.
        target : int
            Index of target variable.

        Returns
        -------
        coef : array-like, shape (n_features)
            Coefficients of predictor variable.
        """
        # Standardize X
        scaler = StandardScaler()
        X_std = scaler.fit_transform(X)

        # Pruning with Adaptive Lasso
        lr = LinearRegression()
        lr.fit(X_std[:, predictors], X_std[:, target])
        weight = np.power(np.abs(lr.coef_), gamma)
        reg = LassoLarsCV()
        reg.fit(X_std[:, predictors] * weight, X_std[:, target])
        pruned_idx = np.abs(reg.coef_ * weight) > 0.0

        # Calculate coefficients of the original scale
        coef = np.zeros(reg.coef_.shape)
        if pruned_idx.sum() > 0:
            lr = LinearRegression()
            pred = np.array(predictors)
            lr.fit(X[:, pred[pruned_idx]], X[:, target])
            coef[pruned_idx] = lr.coef_

        return coef



import numpy as np

from sklearn.utils import check_scalar, check_array, check_random_state
from sklearn.utils.validation import check_is_fitted
from sklearn.model_selection import check_cv
from sklearn.dummy import DummyRegressor
from sklearn.ensemble import RandomForestRegressor


class OutOfSampleCausalTuning():
    """Implementation of Out-of-sample Causal Tuning Algorithm [1]_

    References
    ----------
    .. [1] Biza, Konstantina, Ioannis Tsamardinos, and Sofia Triantafillou.
      "Out-of-Sample Tuning for Causal Discovery."
      IEEE Transactions on Neural Networks and Learning Systems (2022).

    """

    def __init__(self, cv=10, sparsity_penalty=True, thr=0.05, random_state=None):
        """Construct a OutOfSampleCausalTuning.

        Parameters
        ----------
        cv : int, cross-validation generator or an iterable, default=10
            Determines the cross-validation splitting strategy.

        sparsity_penalty : bool
            Whether to enable sparsity penalty.

        thr : float
            Significance threshold of the permutation test of sparsity penalty.

        random_state : None, int or instance of RandomState, default=None
            Controls the permutation test of sparsity penalty when applying
            sparsity penalty.

        Attributes
        ----------
        best_config_index_ : int
            Index of the best configuraion.

        scores_ : dictionary
            Scores of the given configurations.
        """

        self.cv = cv
        self.thr = thr
        self.sparsity_penalty = sparsity_penalty
        self.random_state = random_state

    @property
    def best_config_index_(self):
        check_is_fitted(self)
        return self.best_idx_

    @property
    def scores_(self):
        check_is_fitted(self)
        return self.fit_info_

    def fit(self, data, configs):
        """Search the index of the best configuration given in configs.

        Parameters
        ----------
        data : array-like of shape (n_samples, n_features)
            The input samples.

        configs : list of dictionary
            A list of configurations specifying the causal discovery
            algorithm and its parameters. Each configuration shall have
            ``model`` unless ``all_others`` is True.
            ``model`` specifies the constructor of a causal discovery algorithm
            such as DirectLiNGAM or ICALiNGAM. ``init_kwargs`` and ``fit_kwargs``
            specify the arguments to be given to the constructor and fit of
            the causal discovery algorithm. Set ``all_others`` to True if you
            want to ignore markov blankets when making predictive models
            for each variable.

        Returns
        -------
        self : object
            Instance of fitted estimator.
        """

        # check input args
        data = check_array(data, ensure_min_features=2)
        configs = self._check_configs(configs)

        # check parameters
        cv = check_cv(self.cv)
        thr = check_scalar(
            self.thr,
            "thr",
            float,
            min_val=0,
            max_val=1,
            include_boundaries="neither"
        )
        sp = check_scalar(
            self.sparsity_penalty, "sparsity_penalty", bool
        )
        rs = check_random_state(self.random_state)

        scores = []
        mb_sizes = []
        config_infos = []

        # evaluate each config
        for config in configs:
            config_info = []

            # make model set
            for cv_idx, (train, test) in enumerate(cv.split(data)):
                train_data = data[train]
                test_data = data[test]

                if config["all_others"] is False:
                    graph = self._estimate_graph(train_data, config)
                else:
                    graph = None

                # select explanatories based on the estiamted graph
                exps = self._select_explanatory(graph, data.shape[1])

                # make models of variables
                models = self._make_var_models(train_data, exps, rs)

                # predict
                preds = self._predict_vars(test_data, exps, models)

                cv_info = {
                    "test": test,
                    "graph": graph,
                    "exps": exps,
                    "preds": preds,
                }
                config_info.append(cv_info)

            # evaluate config
            score = self._evaluate_config(data, config_info)

            # calculate average markov blanket size (num of explanatory)
            mb_size = self._get_mb_sizes(config_info)

            scores.append(score)
            mb_sizes.append(mb_size)
            config_infos.append(config_info)

        # select temporal best conf
        best_idx = np.argmax(scores)

        if sp is True:
            best_idx, sp_scores = self._sparsity_penalty(
                data, thr, rs, scores, mb_sizes, config_infos, best_idx
            )
        else:
            sp_scores = None

        self.best_idx_ = best_idx

        fit_info = {}
        fit_info["performance"] = scores
        fit_info["mb_size"] = mb_sizes
        fit_info["sp_score"] = sp_scores
        self.fit_info_ = fit_info

        return self

    def _check_configs(self, configs):
        if not isinstance(configs, list):
            raise TypeError("configs is not a list")

        if len(configs) < 1:
            raise ValueError("configs is empty")

        configs_ = []

        for config_idx, config in enumerate(configs):
            if not isinstance(config, dict):
                raise TypeError(f"configs[{config_idx}] is not a dict")

            # defualt values
            config_ = {
                "all_others": False,
                "model": None,
                "init_kwargs": {},
                "fit_kwargs": {},
            }

            # parse config
            for key, value in config.items():
                if key == "all_others":
                    check_scalar(value, key, bool)
                elif key == "model":
                    if not callable(value):
                        raise TypeError(f"{key} is not callable (config[{config_idx}])")
                elif key == "init_kwargs":
                    if not isinstance(value, dict):
                        raise TypeError(f"{key} is not a dict (config[{config_idx}])")
                elif key == "fit_kwargs":
                    if not isinstance(value, dict):
                        raise TypeError(f"{key} is not a dict (config[{config_idx}])")
                else:
                    raise ValueError(f"{key} is an unknown key (config[{config_idx}])")
                config_[key] = value

            # check
            if config_["all_others"] is False and "model" not in config_.keys():
                raise ValueError(f"model is not specified (config[{config_idx}])")
            elif config_["all_others"] is True:
                config_["model"] = None
                config_["init_kwargs"] = {}
                config_["fit_kwargs"] = {}

            configs_.append(config_)

        return configs_

    def _estimate_graph(self, train_data, config):
        model = config["model"](**config["init_kwargs"])
        model.fit(train_data, **config["fit_kwargs"])

        return model.adjacency_matrix_

    def _get_mb(self, i, graph):
        pa = np.argwhere(~np.isclose(graph[i], 0)).flatten()
        ch = np.argwhere(~np.isclose(graph[:, i], 0)).flatten()

        co_pa = []
        if len(ch) > 0:
            for ch_ in ch:
                co_pa_ = np.argwhere(~np.isclose(graph[ch_], 0)).flatten()
                co_pa.append(co_pa_)
            co_pa = np.concatenate(co_pa)
            co_pa = co_pa[co_pa != i]

        return np.unique([*pa, *ch, *co_pa]).tolist()

    def _select_explanatory(self, graph, n_features):
        exps = []

        for var_idx in range(n_features):
            if graph is not None:
                exp = self._get_mb(var_idx, graph)
            else:
                exp = sorted(set(range(n_features)) - set([var_idx]))
            exps.append(exp)

        return exps

    def _make_var_models(self, train_data, exps, random_state):
        models = []

        for var_idx, exp in enumerate(exps):
            X = train_data[:, exp]
            y = train_data[:, var_idx]

            if len(exp) > 0:
                model = RandomForestRegressor(
                    n_jobs=-1,
                    random_state=random_state,
                )
            else:
                model = DummyRegressor(strategy="mean")

            model.fit(X, y)

            models.append(model)

        return models

    def _predict_vars(self, test_data, exps, models):
        preds = {}

        for var_idx, exp in enumerate(exps):
            X = test_data[:, exp]

            pred = models[var_idx].predict(X)

            preds[var_idx] = pred

        return preds

    def _calc_mutual_info(self, x, x2):
        return -0.5 * np.log(1 - np.corrcoef(x, x2)[0, 1] ** 2)

    def _evaluate_config(self, data, config_info):
        scores = []

        for var_idx in range(data.shape[1]):
            pred = []
            truth = []

            for cv_idx, config_info_ in enumerate(config_info):
                pred.append(config_info_["preds"][var_idx])
                truth.append(data[config_info_["test"], var_idx])

            pred = np.concatenate(pred)
            truth = np.concatenate(truth)

            score = self._calc_mutual_info(truth, pred)

            scores.append(score)

        return np.mean(scores)

    def _get_mb_sizes(self, config_info):
        mb_sizes = []

        for cv_idx, config_info_ in enumerate(config_info):
            for var_idx, exp in enumerate(config_info_["exps"]):
                mb_sizes.append(len(exp))

        return np.mean(mb_sizes)

    def _sparsity_penalty(self, data, thr, rs, scores, mb_sizes, config_infos, best_idx):
        ret = best_idx
        trial_num = 1000

        ps = {}
        for conf_idx, config_info in enumerate(config_infos):
            if conf_idx == best_idx:
                continue

            # permutation test
            t_obs = scores[best_idx] - scores[conf_idx]

            ts = []

            for trial in range(trial_num):
                best_mis = []
                mis = []

                for var_idx in range(data.shape[1]):
                    # concatenate predicted values over CVs
                    best_pred = np.hstack(
                        [cv_info["preds"][var_idx] for cv_info in config_infos[best_idx]]
                    )

                    # concatenate predicted values over CVs
                    pred = np.hstack([cv_info["preds"][var_idx] for cv_info in config_info])

                    # concatenate true values over CVs
                    test_indices = np.hstack([cv_info["test"] for cv_info in config_info])
                    truth = np.hstack(data[test_indices, var_idx])

                    # swap
                    best_pred_pm = rs.permutation(best_pred)
                    pred_pm = rs.permutation(pred)

                    # mutual info with original data
                    best_mi = self._calc_mutual_info(truth, best_pred_pm)
                    mi = self._calc_mutual_info(truth, pred_pm)

                    best_mis.append(best_mi)
                    mis.append(mi)

                t = np.mean(best_mis) - np.mean(mis)
                ts.append(t)

            p = sum(map(lambda t: t > t_obs, ts)) / trial_num
            ps[conf_idx] = p

        # select configuration
        matches = [k for k, v in ps.items() if v > thr]

        if len(matches) > 0:
            matches = [best_idx] + matches

            mbs = [mb_sizes[idx] for idx in matches]
            ret = matches[np.argmin(mbs)]

        return ret, ps



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""

from .oct import OutOfSampleCausalTuning
from .cdg import CausalDataGenerator

__all__ = [
    "OutOfSampleCausalTuning",
    "CausalDataGenerator",
]



import os

import numpy as np
import pandas as pd

from lingam.longitudinal_lingam import LongitudinalLiNGAM


def test_fit_success():
    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=1000)
    x1 = 0.7*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 0.3*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000) + 0.5*x2
    x1 = 0.3*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000) + 0.5*x3
    x3 = 0.7*x1 + np.random.uniform(size=1000)
    X2 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000) + 0.5*x2
    x1 = 0.5*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000) + 0.5*x3
    x3 = 0.5*x1 + np.random.uniform(size=1000)
    X3 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    X_list = np.empty((3, 1000, 4))
    X_list[0] = X1
    X_list[1] = X2
    X_list[2] = X3

    model = LongitudinalLiNGAM()
    model.fit(X_list)

    # check causal ordering
    cos = model.causal_orders_
    for co in cos[1:]:
        assert co.index(0) < co.index(1) < co.index(3)

    # check B(t,t)
    B_t = model.adjacency_matrices_[1, 0] # B(1,1)
    assert B_t[1, 0] > 0.1 and B_t[3, 1] > 0.5
    B_t[1, 0] = 0.0
    B_t[3, 1] = 0.0
    assert np.sum(B_t) < 0.1

    B_t = model.adjacency_matrices_[2, 0] # B(2,2)
    assert B_t[1, 0] > 0.3 and B_t[3, 1] > 0.3
    B_t[1, 0] = 0.0
    B_t[3, 1] = 0.0
    assert np.sum(B_t) < 0.1

    # check B(t,t-τ)
    B_tau = model.adjacency_matrices_[1, 1] # B(1,0)
    assert B_tau[0, 2] > 0.3 and B_tau[2, 3] > 0.3

    B_tau = model.adjacency_matrices_[1, 1] # B(2,1)
    assert B_tau[0, 2] > 0.3 and B_tau[2, 3] > 0.3

    # fit by list
    X_list = [X1, X2, X3]
    model = LongitudinalLiNGAM()
    model.fit(X_list)

    p_values = model.get_error_independence_p_values()
    resid = model.residuals_

    # prior knowledge
    pk = np.ones((3, 2, 4, 4)) * -1
    model = LongitudinalLiNGAM(prior_knowledge=pk)
    model.fit(X_list)

def test_fit_invalid_data():
    # Different features
    x0 = np.random.uniform(size=1000)
    x1 = 0.7*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 0.3*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000) + 0.5*x2
    x1 = 0.3*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000) + 0.5*x3
    X2 = pd.DataFrame(np.array([x0, x1, x2]).T,
                      columns=['x0', 'x1', 'x2'])

    x0 = np.random.uniform(size=1000) + 0.5*x2
    x1 = 0.5*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000) + 0.5*x3
    x3 = 0.5*x1 + np.random.uniform(size=1000)
    X3 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    X_list = [X1, X2, X3]

    try:
        model = LongitudinalLiNGAM()
        model.fit(X_list)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Not list data
    X = 1
    try:
        model = LongitudinalLiNGAM()
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include not-array data
    x0 = np.random.uniform(size=1000)
    x1 = 2.0*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    X_list = [X1, 1]

    try:
        model = LongitudinalLiNGAM()
        model.fit(X_list)
    except ValueError:
        pass
    else:
        raise AssertionError

    # List items
    X_list = [X1]

    try:
        model = LongitudinalLiNGAM()
        model.fit(X_list)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include non-numeric data
    x0 = np.random.uniform(size=1000)
    x1 = 0.7*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 0.3*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000) + 0.5*x2
    x1 = np.array(['X']*1000) # <== non-numeric
    x2 = np.random.uniform(size=1000) + 0.5*x3
    x3 = np.array(['X']*1000) # <== non-numeric
    X2 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000) + 0.5*x2
    x1 = 0.5*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 0.5*x1 + np.random.uniform(size=1000)
    X3 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    X_list = [X1, X2, X3]

    try:
        model = LongitudinalLiNGAM()
        model.fit(X_list)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include NaN values
    x0 = np.random.uniform(size=1000)
    x1 = 0.7*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 0.3*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000) + 0.5*x2
    x1 = 0.3*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000) + 0.5*x3
    x3 = 0.7*x1 + np.random.uniform(size=1000)
    X2 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])
    X2.iloc[100, 0] = np.nan # set nan

    x0 = np.random.uniform(size=1000) + 0.5*x2
    x1 = 0.5*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000) + 0.5*x3
    x3 = 0.5*x1 + np.random.uniform(size=1000)
    X3 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    X_list = [X1, X2, X3]

    try:
        model = LongitudinalLiNGAM()
        model.fit(X_list)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include infinite values
    x0 = np.random.uniform(size=1000)
    x1 = 0.7*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 0.3*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000) + 0.5*x2
    x1 = 0.3*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000) + 0.5*x3
    x3 = 0.7*x1 + np.random.uniform(size=1000)
    X2 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])
    X2.iloc[100, 0] = np.inf # set inf

    x0 = np.random.uniform(size=1000) + 0.5*x2
    x1 = 0.5*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000) + 0.5*x3
    x3 = 0.5*x1 + np.random.uniform(size=1000)
    X3 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    X_list = [X1, X2, X3]

    try:
        model = LongitudinalLiNGAM()
        model.fit(X_list)
    except ValueError:
        pass
    else:
        raise AssertionError

    # prior knowledge
    pk = np.ones((3, 2, 4, 4)) * -1
    try:
        # pk.shape[1] != n_lags + 1
        model = LongitudinalLiNGAM(n_lags=10, prior_knowledge=pk)
        model.fit(X_list)
    except ValueError:
        pass
    else:
        raise AssertionError

def test_bootstrap_success():
    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=1000)
    x1 = 0.7*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 0.3*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000) + 0.5*x2
    x1 = 0.3*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000) + 0.5*x3
    x3 = 0.7*x1 + np.random.uniform(size=1000)
    X2 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000) + 0.5*x2
    x1 = 0.5*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000) + 0.5*x3
    x3 = 0.5*x1 + np.random.uniform(size=1000)
    X3 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    X_list = np.empty((3, 1000, 4))
    X_list[0] = X1
    X_list[1] = X2
    X_list[2] = X3

    model = LongitudinalLiNGAM()
    model.bootstrap(X_list, n_sampling=3)

    # fit by list
    X_list = [X1, X2, X3]
    model = LongitudinalLiNGAM()
    result = model.bootstrap(X_list, n_sampling=3)

    result.adjacency_matrices_
    result.total_effects_

    # No argument
    cdc = result.get_causal_direction_counts()

    # n_directions=2
    cdc = result.get_causal_direction_counts(n_directions=2)

    # min_causal_effect=0.2
    cdc = result.get_causal_direction_counts(min_causal_effect=0.2)

    # split_by_causal_effect_sign=True
    cdc = result.get_causal_direction_counts(split_by_causal_effect_sign=True)

    # No argument
    dagc = result.get_directed_acyclic_graph_counts()

    # n_dags=2
    dagc = result.get_directed_acyclic_graph_counts(n_dags=2)

    # min_causal_effect=0.6
    dagc = result.get_directed_acyclic_graph_counts(min_causal_effect=0.6)

    # split_by_causal_effect_sign=True
    dagc = result.get_directed_acyclic_graph_counts(split_by_causal_effect_sign=True)

    # get_probabilities
    probs = result.get_probabilities()

    # get_probabilities
    probs = result.get_probabilities(min_causal_effect=0.6)

    # get_total_causal_effects
    ce = result.get_total_causal_effects()

    # get_total_causal_effects
    ce = result.get_total_causal_effects(min_causal_effect=0.6)

    # get_paths
    result.get_paths(0, 1, 1, 2)


def test_bootstrap_invalid_data():
    # Different features
    x0 = np.random.uniform(size=1000)
    x1 = 0.7*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 0.3*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000) + 0.5*x2
    x1 = 0.3*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000) + 0.5*x3
    X2 = pd.DataFrame(np.array([x0, x1, x2]).T,
                      columns=['x0', 'x1', 'x2'])

    x0 = np.random.uniform(size=1000) + 0.5*x2
    x1 = 0.5*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000) + 0.5*x3
    x3 = 0.5*x1 + np.random.uniform(size=1000)
    X3 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    X_list = [X1, X2, X3]

    try:
        model = LongitudinalLiNGAM()
        model.bootstrap(X_list, n_sampling=3)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Not list data
    X = 1
    try:
        model = LongitudinalLiNGAM()
        model.bootstrap(X, n_sampling=3)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include not-array data
    x0 = np.random.uniform(size=1000)
    x1 = 2.0*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    X_list = [X1, 1]

    try:
        model = LongitudinalLiNGAM()
        model.bootstrap(X_list, n_sampling=3)
    except ValueError:
        pass
    else:
        raise AssertionError

    # List items
    X_list = [X1]

    try:
        model = LongitudinalLiNGAM()
        model.bootstrap(X_list, n_sampling=3)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include non-numeric data
    x0 = np.random.uniform(size=1000)
    x1 = 0.7*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 0.3*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000) + 0.5*x2
    x1 = np.array(['X']*1000) # <== non-numeric
    x2 = np.random.uniform(size=1000) + 0.5*x3
    x3 = np.array(['X']*1000) # <== non-numeric
    X2 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000) + 0.5*x2
    x1 = 0.5*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 0.5*x1 + np.random.uniform(size=1000)
    X3 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    X_list = [X1, X2, X3]

    try:
        model = LongitudinalLiNGAM()
        model.bootstrap(X_list, n_sampling=3)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include NaN values
    x0 = np.random.uniform(size=1000)
    x1 = 0.7*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 0.3*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000) + 0.5*x2
    x1 = 0.3*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000) + 0.5*x3
    x3 = 0.7*x1 + np.random.uniform(size=1000)
    X2 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])
    X2.iloc[100, 0] = np.nan # set nan

    x0 = np.random.uniform(size=1000) + 0.5*x2
    x1 = 0.5*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000) + 0.5*x3
    x3 = 0.5*x1 + np.random.uniform(size=1000)
    X3 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    X_list = [X1, X2, X3]

    try:
        model = LongitudinalLiNGAM()
        model.bootstrap(X_list, n_sampling=3)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include infinite values
    x0 = np.random.uniform(size=1000)
    x1 = 0.7*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 0.3*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000) + 0.5*x2
    x1 = 0.3*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000) + 0.5*x3
    x3 = 0.7*x1 + np.random.uniform(size=1000)
    X2 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])
    X2.iloc[100, 0] = np.inf # set inf

    x0 = np.random.uniform(size=1000) + 0.5*x2
    x1 = 0.5*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000) + 0.5*x3
    x3 = 0.5*x1 + np.random.uniform(size=1000)
    X3 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    X_list = [X1, X2, X3]

    try:
        model = LongitudinalLiNGAM()
        model.bootstrap(X_list, n_sampling=3)
    except ValueError:
        pass
    else:
        raise AssertionError

    x0 = np.random.uniform(size=1000)
    x1 = np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T, columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000)
    x1 = np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = np.random.uniform(size=1000)
    X2 = pd.DataFrame(np.array([x0, x1, x2, x3]).T, columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000)
    x1 = np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = np.random.uniform(size=1000)
    X3 = pd.DataFrame(np.array([x0, x1, x2, x3]).T, columns=['x0', 'x1', 'x2', 'x3'])
    
    X_list = [X1, X2, X3]

    # Invalid argument: bootstrap(n_sampling=-1)
    model = LongitudinalLiNGAM()
    try:
        result = model.bootstrap(X_list, n_sampling=-1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_causal_direction_counts(n_directions=-1)
    model = LongitudinalLiNGAM()
    result = model.bootstrap(X_list, n_sampling=5)
    try:
        result.get_causal_direction_counts(n_directions=-1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_causal_direction_counts(n_directions='3')
    try:
        result.get_causal_direction_counts(n_directions='3')
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_causal_direction_counts(min_causal_effect=-1.0)
    try:
        result.get_causal_direction_counts(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_directed_acyclic_graph_counts(n_dags=-1)
    try:
        result.get_directed_acyclic_graph_counts(n_dags=-1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_directed_acyclic_graph_counts(n_dags='3')
    try:
        result.get_directed_acyclic_graph_counts(n_dags='3')
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_directed_acyclic_graph_counts(min_causal_effect=-1.0)
    try:
        result.get_directed_acyclic_graph_counts(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_probabilities(min_causal_effect=-1.0)
    try:
        result.get_probabilities(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_total_causal_effects(min_causal_effect=-1.0)
    try:
        result.get_total_causal_effects(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_paths(min_causal_effect<0)
    try:
        result.get_paths(0, 1, 1, 2, -1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_paths(to_t < from_t)
    try:
        result.get_paths(0, 1, 1, 0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_paths(to_t < from_t)
    try:
        result.get_paths(0, 1, 1, 0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_paths(same variable)
    try:
        result.get_paths(0, 0, 1, 1)
    except ValueError:
        pass
    else:
        raise AssertionError

def test_estimate_total_effect_invalid():
    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=1000)
    x1 = 0.7*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 0.3*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000) + 0.5*x2
    x1 = 0.3*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000) + 0.5*x3
    x3 = 0.7*x1 + np.random.uniform(size=1000)
    X2 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000) + 0.5*x2
    x1 = 0.5*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000) + 0.5*x3
    x3 = 0.5*x1 + np.random.uniform(size=1000)
    X3 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    X_list = np.empty((3, 1000, 4))
    X_list[0] = X1
    X_list[1] = X2
    X_list[2] = X3

    model = LongitudinalLiNGAM()
    model.fit(X_list)
    model._causal_orders[0] = [0, 1, 2, 3]
    model._causal_orders[1] = [0, 1, 2, 3]
    model._causal_orders[2] = [0, 1, 2, 3]

    # warning
    model.estimate_total_effect(X_list, 2, 1, 2, 0)
    model.estimate_total_effect(X_list, 2, 1, 1, 1)



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""

from itertools import product

import networkx as nx
import numpy as np
import pandas as pd
import scipy.optimize as sopt
from scipy.special import expit as sigmoid
from sklearn.linear_model import (LinearRegression, LogisticRegression,
                                  PoissonRegressor)
from sklearn.metrics import log_loss

import lingam.utils as ut


class LiM:
    """Implementation of LiM Algorithm [1]_

    References
    ----------
    .. [1] Zeng Y, Shimizu S, Matsui H, et al. Causal discovery for
       linear mixed data[C]//Conference on Causal Learning and Rea-
       soning. PMLR, 2022: 994-1009.
    """

    def __init__(
        self,
        lambda1=0.1,
        loss_type="mixed",
        max_iter=150,
        h_tol=1e-8,
        rho_max=1e16,
        w_threshold=0.1,
    ):
        """Construct a LiM model.

        Parameters
        ----------
        lambda1 : float, optional (default=0.1)
             L1 penalty parameter.
        loss_type : str, (default='mixed')
             Type of distribution of the noise.
        max_iter : int, (default=150)
             Maximum number of dual ascent steps.
        h_tol : float, (default=1e-8)
             Tolerance parameter of the acyclicity constraint.
        rho_max : float, (default=1e+16)
             Maximum value of the regularization parameter rho.
        w_threshold : float (default=0.1)
             Drop edge if the weight btw. variables is less than w_threshold.
        """
        self._lambda1 = lambda1
        self._loss_type = loss_type
        self._max_iter = max_iter
        self._h_tol = h_tol
        self._rho_max = rho_max
        self._w_threshold = w_threshold
        self._adjacency_matrix = None

    def fit(self, X, dis_con, only_global=False, is_poisson=False):
        """Fit the model to X with mixed data.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of observed variables.
        dis_con: array-like, shape (1, n_features)
            Indicators of discrete or continuous variables, where "1"
            indicates a continuous variable, while "0" a discrete
            variable.
        only_global: boolean, optional (default=False)
            If True, then the method will only perform the global optimization
            to estimate the causal structure, without the local search phase.
        is_poisson: boolean, optional (default=False)
            If True, then the method will use poisson regression model to compute
            the log-likelihood in the local search phase.

        Returns
        -------
        self : object
            Returns the instance of self.
        """

        W_min_lss = self._estimate_LiM(X, dis_con, only_global, is_poisson)
        self._adjacency_matrix = W_min_lss
        return self._adjacency_matrix

    def _estimate_LiM(self, X, dis_con, only_global, is_poisson):
        """Estimate the adjacency matrix btw. mixed variables"""

        def _loss(W):
            """Evaluate value and gradient of loss."""
            M = X @ W
            if self._loss_type == "logistic":
                loss = 1.0 / X.shape[0] * (np.logaddexp(0, M) - X * M).sum()
                G_loss = 1.0 / X.shape[0] * X.T @ (sigmoid(M) - X)
            elif self._loss_type == "poisson":
                loss = -np.sum(X.T @ M) + np.exp(M).sum()
                for j in range(X.shape[0]):
                    for k in range(X.shape[1]):
                        loss += np.log(_factorial(X[j][k]))
                loss = 1.0 / X.shape[0] * loss
                G_loss = 1.0 / X.shape[0] * (-X.T @ X + X.T @ np.exp(M))
            elif self._loss_type == "laplace":
                R = X - M
                loss = -1.0 / X.shape[0] * np.sum(-np.log(np.cosh(R)))
                G_loss = 1.0 / X.shape[0] * X.T @ np.tanh(R)
            elif self._loss_type == "mixed":
                R = X - M
                a1, a2 = 1, 1
                loss_dis = (
                    (np.logaddexp(0, M) - X * M) * np.absolute(dis_con - 1)
                ).sum()
                loss_con = np.sum(-np.log(np.cosh(R)) * dis_con)
                loss = -1.0 / X.shape[0] * (a1 * loss_dis + a2 * loss_con)
                W_dis = np.zeros([d, d])
                W_con = np.zeros([d, d])
                for ii in np.where(dis_con[0, :] == 0):
                    W_dis[[ii], :] = 1
                    W_dis[:, [ii]] = 1
                for jj in np.where(dis_con[0, :] == 1):
                    W_con[[jj], :] = 1
                    W_con[:, [jj]] = 1
                G_dis = X.T @ (sigmoid(M) - X) * W_dis
                G_con = -X.T @ np.tanh(R) * W_con
                G_loss = 1.0 / X.shape[0] * (G_dis + G_con)
            elif self._loss_type == "mixed_dag":
                dag = nx.DiGraph(W)
                lingam_data = np.transpose(X)
                df = pd.DataFrame(X)
                is_continuous = dis_con[0, :].astype(bool)
                is_discrete = np.invert(is_continuous)
                loss = -_bic_scoring(
                    dag, is_discrete, is_poisson, df, lingam_data
                )  # lingam_data:dims*samples
                R = X - M
                G_loss = 1.0 / X.shape[0] * (X.T @ (sigmoid(M) - X) - X.T @ np.tanh(R))
            else:
                raise ValueError("unknown loss type")
            return loss, G_loss

        def _bic_scoring(dag: nx.DiGraph, is_discrete, is_poisson, df, lingam_data):
            """Evaluate value of loss given the DAG."""
            sample_size = df.shape[0]
            K = dag.number_of_edges() + dag.number_of_nodes()
            # K = dag.number_of_edges() + np.sum(is_discrete)
            penalty = np.log(sample_size) / 2 * K
            total_score = 0.0 - penalty

            for i in dag.nodes:
                parents_i = [j for j in dag.predecessors(i)]

                # 離散 logistic
                if is_discrete[i] and is_poisson == False:
                    if not parents_i:
                        # Bernoulli binary variable, likelihood using Bernoulli model with MLE parameters
                        frequency_table = df[i].value_counts()
                        likekihood_bernoulli = 0.0
                        for count_k in frequency_table:
                            likekihood_bernoulli += count_k * (
                                np.log(count_k) - np.log(frequency_table.sum())
                            )
                        total_score += likekihood_bernoulli

                    elif parents_i:
                        # Logistic binary variable, likelihood using logistic regression model.
                        X = df[parents_i]
                        y = df[i]
                        logistic = LogisticRegression(
                            solver="lbfgs"
                        )  # or solver='liblinear'
                        logistic.fit(X, y)
                        predict_prob = logistic.predict_proba(X)

                        # Negative cross-entropy loss a.k.a log-likelihood
                        likekihood_logistic = -log_loss(
                            y_true=y, y_pred=predict_prob, normalize=False
                        )
                        total_score += likekihood_logistic
                    pass

                # 離散 poisson
                elif is_discrete[i] and is_poisson == True:
                    if not parents_i:
                        # Bernoulli count variable, likelihood using Bernoulli model with MLE parameters
                        frequency_table = df[i].value_counts()
                        likekihood_bernoulli = 0.0
                        for count_k in frequency_table:
                            likekihood_bernoulli += count_k * (
                                np.log(count_k) - np.log(frequency_table.sum())
                            )
                        total_score += likekihood_bernoulli

                    elif parents_i:
                        # Count variable, likelihood using poisson regression model.
                        for iii in range(len(parents_i)):
                            X = lingam_data[parents_i[iii]]
                            X = X.reshape(-1, 1)
                            y = lingam_data[i]
                            poisson = PoissonRegressor()
                            poisson.fit(X, y)
                            beta = poisson.coef_
                            # compute likelihood
                            likekihood_poisson = -np.sum(
                                y * lingam_data[parents_i[iii]] * beta
                            ) + np.sum(np.exp(lingam_data[parents_i[iii]] * beta))
                            for j in range(len(y)):
                                likekihood_poisson += np.log(_factorial(y[j]))
                        total_score += likekihood_poisson
                        # # or we can compute likelihood via nn
                        # pnllloss = nn.PoissonNLLLoss()
                        # log_input = np.log(lingam_data[parents_i] * beta)
                        # target = y # torch.randn(5, 2)
                        # output = pnllloss(log_input, target)
                        # likekihood_poisson = output.item()
                        # total_score += likekihood_poisson

                # 連続
                elif not is_discrete[i]:
                    b_i = np.zeros(dag.number_of_nodes())

                    if parents_i:  # estimate b_i
                        X = df[parents_i]
                        y = df[i]
                        lr = LinearRegression(fit_intercept=True)  # because zero means
                        lr.fit(X, y)
                        for index, j in enumerate(parents_i):
                            b_i[j] = lr.coef_[index]

                        bi_0 = lr.intercept_
                    else:
                        bi_0 = np.mean(df[i])

                    likekihood_lingam = ut.likelihood_i(lingam_data, i, b_i, bi_0)
                    total_score += likekihood_lingam
                    pass
            return total_score

        def _h(W):
            """Evaluate value and gradient of acyclicity constraint."""
            # E = slin.expm(W * W)  # (Zheng et al. 2018)
            # h = np.trace(E) - d
            # A different formulation, slightly faster at the cost of numerical stability
            M = np.eye(d) + W * W / d  # (Yu et al. 2019)
            G_h = np.linalg.matrix_power(M, d - 1)
            h = (G_h.T * M).sum() - d
            return h, G_h

        def _adj(w):
            """Convert doubled variables ([2 d^2] array) back to original variables ([d, d] matrix)."""
            return (w[: d * d] - w[d * d :]).reshape([d, d])

        def _func(w):
            """Evaluate value and gradient of augmented Lagrangian for doubled variables ([2 d^2] array)."""
            W = _adj(w)
            loss, G_loss = _loss(W)
            h, G_h = _h(W)
            obj = loss + 0.5 * rho * h * h + alpha * h + self._lambda1 * w.sum()
            # G_smooth = G_loss + (rho * h + alpha) * G_h # Zheng Xun Dag 2018
            G_smooth = G_loss + (rho * h + alpha) * G_h.T * W * 2  # 2019
            g_obj = np.concatenate(
                (G_smooth + self._lambda1, -G_smooth + self._lambda1), axis=None
            )
            return obj, g_obj

        def _factorial(y):
            if not isinstance(y, int):
                y = round(y)
            if y == 0 or y == 1 or y < 0:
                return 1
            else:
                return y * _factorial(y - 1)

        n, d = X.shape
        w_est, rho, alpha, h = (
            np.random.random(2 * d * d),
            1.0,
            0.0,
            np.inf,
        )  # double w_est into (w_pos, w_neg)
        bnds = [
            (0, 0) if i == j else (0, None)
            for _ in range(2)
            for i in range(d)
            for j in range(d)
        ]

        # if self._loss_type == 'l2':
        #     X = X - np.mean(X, axis=0, keepdims=True)
        for _ in range(self._max_iter):
            w_new, h_new = None, None
            while rho < self._rho_max:
                sol = sopt.minimize(
                    _func, w_est, method="L-BFGS-B", jac=True, bounds=bnds
                )
                # print('--- One iteration passed.....', sol.fun)
                w_new = sol.x
                h_new, _ = _h(_adj(w_new))
                if h_new >= 0.25 * h:
                    rho *= 10
                else:
                    break
            w_est, h = w_new, h_new
            alpha += rho * h
            # print('------- rho  is:', rho)
            if h <= self._h_tol and h != 0:
                break
            if rho >= self._rho_max * 1e-6 and h > 1e05:  # avoid the full graph
                w_est = np.random.random(2 * d * d)
                rho = 1.0
            elif rho >= self._rho_max:
                break
            if np.sum(np.absolute(_adj(w_est))) < 0.09:  # avoid the zero matrix
                w_est = np.random.random(2 * d * d)
        W_est = _adj(w_est)
        W_est[np.abs(W_est) < self._w_threshold] = 0
        print("W_est (without the 2nd phase) is: \n", W_est)

        if not only_global:
            self._loss_type = "mixed_dag"
            aa, aaa = _loss(W_est)  # loss
            I = np.where(W_est != 0)
            # check directions
            W_min_lss = np.copy(W_est)
            candi_setting = list(
                product(range(2), repeat=len(I[0]))
            )  # 1:reverse the direction； 0:unchanged
            for candi_setting_i in range(1, len(candi_setting)):
                W_tmp = np.copy(W_est)
                for iii in range(len(I[0])):  # transform to W_tmp
                    if candi_setting[candi_setting_i][iii] == 1:
                        W_tmp[I[0][iii], I[1][iii]] = 0
                        W_tmp[I[1][iii], I[0][iii]] = 1
                lss, lss_G = _loss(W_tmp)
                if lss < aa and _h(W_tmp)[0] < self._h_tol:
                    W_min_lss = np.copy(W_tmp)
                    aa = lss

            # prune process
            if d > 2 and len(I[0]) > (d - 1):  # > min_edge
                W0 = np.copy(W_min_lss)
                I_delete = np.where(W_min_lss != 0)
                for delete_i in range(len(I_delete[0])):
                    W_tmp = np.copy(W0)
                    W_tmp[I_delete[0][delete_i], I_delete[1][delete_i]] = 0
                    lss, lss_G = _loss(W_tmp)
                    if lss < aa and _h(W_tmp)[0] < self._h_tol:
                        W_min_lss = np.copy(W_tmp)
                        aa = lss
                #
                if not np.all(
                    W_est.astype(bool) == W_min_lss.astype(bool)
                ):  # if they are different
                    W0 = np.copy(W_est)
                    for delete_i in range(len(I[0])):
                        W_tmp = np.copy(W0)
                        W_tmp[I[0][delete_i], I[1][delete_i]] = 0
                        lss, lss_G = _loss(W_tmp)
                        if lss < aa and _h(W_tmp)[0] < self._h_tol:
                            W_min_lss = np.copy(W_tmp)
                            aa = lss
            # add process
            if d > 2 and len(I[0]) < (d * (d - 1) / 2):
                W0 = np.copy(W_min_lss)
                W_edges = W0 + W0.T + np.eye(d)
                I_add = np.where(W_edges == 0)  # add undirected edges' indices
                for add_i in range(len(I_add[0])):
                    W_tmp = np.copy(W0)
                    W_tmp[I_add[0][add_i], I_add[1][add_i]] = 1
                    lss, lss_G = _loss(W_tmp)
                    if lss < aa and _h(W_tmp)[0] < self._h_tol:
                        W_min_lss = np.copy(W_tmp)
                        aa = lss
                #
                if not np.all(
                    W_est.astype(bool) == W_min_lss.astype(bool)
                ):  # if they are different
                    W0 = np.copy(W_est)
                    W_edges = W0 + W0.T + np.eye(d)
                    I_add = np.where(W_edges == 0)  # add undirected edges' indices
                    for add_i in range(len(I_add[0])):
                        W_tmp = np.copy(W0)
                        W_tmp[I_add[0][add_i], I_add[1][add_i]] = 1
                        lss, lss_G = _loss(W_tmp)
                        if lss < aa and _h(W_tmp)[0] < self._h_tol:
                            W_min_lss = np.copy(W_tmp)
                            aa = lss
            # print('W_min_lss is:\n', W_min_lss)
            # print('W_true is:\n', W_true)

        else:
            W_min_lss = np.copy(W_est)

        return W_min_lss

    @property
    def adjacency_matrix_(self):
        """Estimated adjacency matrix between mixed variables.

        Returns
        -------
        adjacency_matrix_ : array-like, shape (n_features, n_features)
            The adjacency matrix of variables, where
            ``n_features`` is the number of observed variables.
        """
        return self._adjacency_matrix



import pytest

import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from sklearn.linear_model import LogisticRegression
from scipy.special import expit

import os
import sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from lingam.causal_based_simulator import CausalBasedSimulator
from lingam.causal_based_simulator import CBSILiNGAM
from lingam.causal_based_simulator import CBSIUnobsCommonCauseLiNGAM
from lingam.causal_based_simulator import CBSITimeSeriesLiNGAM


@pytest.fixture
def init():
    return lambda : np.random.seed(0)


@pytest.fixture
def test_data():
    np.random.seed(0)

    N = 1000
    causal_graph = np.array([
        [0.0, 0.0, 0.0, 3.0, 0.0, 0.0],
        [3.0, 0.0, 2.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 6.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [8.0, 0.0, -1.0, 0.0, 0.0, 0.0],
        [4.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    ])

    def _is_correct_co(causal_order):
        truth = [{"3"}, {"0", "2"}, {"1", "4", "5"}]
        target = [
            {*causal_order[:1]},
            {*causal_order[1:3]},
            {*causal_order[3:]}
        ]
        return truth == target

    e = np.random.uniform(-np.sqrt(3), np.sqrt(3), size=(len(causal_graph), N))
    X = np.linalg.pinv(np.eye(len(causal_graph)) - causal_graph) @ e
    X = X.T

    return X, causal_graph, _is_correct_co


@pytest.fixture
def test_data_unobs():
    np.random.seed(0)

    N = 1000
    causal_graph = np.array([
        [0.0, 0.0, 0.0, 3.0, 0.0, 0.0],
        [3.0, 0.0, 2.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 6.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [8.0, 0.0, -1.0, 0.0, 0.0, 0.0],
        [4.0, 0.0, 0.0, 0.0, 0.0, 0.0]])

    e = np.random.uniform(-np.sqrt(3), np.sqrt(3), size=(len(causal_graph), N))
    X = np.linalg.pinv(np.eye(len(causal_graph)) - causal_graph) @ e
    X = X.T

    delete_index = 3
    causal_graph = np.delete(causal_graph, delete_index, axis=0)
    causal_graph = np.delete(causal_graph, delete_index, axis=1)
    X = np.delete(X, delete_index, axis=1)
    causal_graph[0, 2] = np.nan
    causal_graph[2, 0] = np.nan

    return X, causal_graph


@pytest.fixture
def test_data_ts():
    np.random.seed(0)

    N = 1000
    causal_graph = np.array([[
        [0, -0.12, 0, 0, 0],
        [0, 0, 0, 0, 0],
        [-0.41, 0.01, 0, -0.02, 0],
        [0.04, -0.22, 0, 0, 0],
        [0.15, 0, -0.03, 0, 0],
    ], [
        [-0.32, 0, 0.12, 0.32, 0],
        [0, -0.35, -0.1, -0.46, 0.4],
        [0, 0, 0.37, 0, 0.46],
        [-0.38, -0.1, -0.24, 0, -0.13],
        [0, 0, 0, 0, 0],
    ]])

    def _is_correct_co(causal_order):
        truth = [
            {"0[t-1]", "1[t-1]", "2[t-1]", "3[t-1]", "4[t-1]"},
            {"1[t]"}, {"0[t]"}, {"3[t]"}, {"2[t]"}, {"4[t]"},
        ]
        target = [
            {*causal_order[:5]},
            {*causal_order[5:6]},
            {*causal_order[6:7]},
            {*causal_order[7:8]},
            {*causal_order[8:9]},
            {*causal_order[9:10]},
        ]
        return truth == target

    def _x_t(X, e, causal_graph):
        _, n_features, _ = causal_graph.shape
        term = np.linalg.pinv(np.eye(n_features) - causal_graph[0])
        term2 = np.hstack(causal_graph[1:]) @ np.hstack(X[:, ::-1][:, :len(causal_graph) - 1]).reshape(n_features, 1) + e
        return term @ term2

    n_lags = len(causal_graph) - 1
    size = N + n_lags

    errors = np.random.uniform(-np.sqrt(3), np.sqrt(3), size=(causal_graph.shape[1], size))

    # initial data
    X = errors[:, :n_lags]

    for t in range(n_lags, size):
        new_data = _x_t(X[:, :t], errors[:, [t]], causal_graph)
        X = np.append(X, new_data, axis=1)
    X = X[:, n_lags:].T

    return X, causal_graph, _is_correct_co


@pytest.fixture
def test_data_discrete():
    np.random.seed(0)

    N = 1000
    causal_graph = np.array([
        [0.0, 0.0, 0.0, 3.0, 0.0, 0.0],
        [3.0, 0.0, 2.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 6.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [8.0, 0.0, -1.0, 0.0, 0.0, 0.0],
        [4.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    ])
    discrete_indices = [0, 1]
    causal_order = [3, 0, 2, 1, 4, 5]

    def _is_correct_co(causal_order):
        truth = [{"3"}, {"0", "2"}, {"1", "4", "5"}]
        target = [
            {*causal_order[:1]},
            {*causal_order[1:3]},
            {*causal_order[3:]}
        ]
        return truth == target

    def e():
        return np.random.uniform(-np.sqrt(3), np.sqrt(3), size=N)

    X = np.zeros((len(causal_graph), N))

    for co in causal_order:
        coefs = causal_graph[co]
        is_exog = np.all(np.isclose(coefs, 0))
        if is_exog:
            X[co] = e()
            continue

        X[co] = coefs @ X + e()
        if co in discrete_indices:
            X[co] = (expit(X[co]) > np.random.uniform(size=N)).astype(int)

    X = X.T

    is_discrete = np.array([False for _ in range(len(causal_graph))])
    is_discrete[discrete_indices] = True

    return X, causal_graph, _is_correct_co, is_discrete


def test_cbs_success(test_data, test_data_unobs, test_data_ts, test_data_discrete):
    models = {"0": LinearRegression()}
    models_d = {"0": LogisticRegression()}
    models_ts = {"0[t]": LinearRegression()}
    changing_models = {"0": {"parent_names": []}}
    changing_models_ts = {"0[t]": {"parent_names": []}}
    changing_exog = {"2": np.random.uniform(-10, 10, size=1000)}
    changing_exog_ts = {"2[t]": np.random.uniform(-10, 10, size=999)}

    sim = CausalBasedSimulator()

    # normal data
    X, causal_graph, _ = test_data
    sim.train(X, causal_graph)
    sim.train(X, causal_graph, models=models)
    sim.run()
    sim.run(changing_models=changing_models, changing_exog=changing_exog)

    # unobserved
    X, causal_graph = test_data_unobs
    sim.train(X, causal_graph, cd_algo_name="BottomUpParceLiNGAM")
    sim.train(X, causal_graph, cd_algo_name="BottomUpParceLiNGAM", models=models)
    sim.run()
    sim.run(changing_models=changing_models, changing_exog=changing_exog)

    # time series
    X, causal_graph, _ = test_data_ts
    sim.train(X, causal_graph, cd_algo_name="VARLiNGAM")
    sim.train(X, causal_graph, cd_algo_name="VARLiNGAM", models=models_ts)
    sim.run()
    sim.run(changing_models=changing_models_ts, changing_exog=changing_exog_ts)

    # discrete
    X, causal_graph, _, is_discrete = test_data_discrete
    sim.train(X, causal_graph, is_discrete=is_discrete)
    sim.train(X, causal_graph, is_discrete=is_discrete, models=models_d)
    sim.run()
    sim.run(changing_models=changing_models, changing_exog=changing_exog)


def test_cbs_exception(test_data):
    X, causal_graph, _ = test_data

    sim = CausalBasedSimulator()

    # cd_algo_name
    try:
        sim.train(X, causal_graph, cd_algo_name=1234)
    except Exception:
        pass
    else:
        raise AssertionError

    try:
        sim.train(X, causal_graph, cd_algo_name="UnknownAlgoName")
    except Exception:
        pass
    else:
        raise AssertionError

    # changing_models makes causal_graph cyclic
    changing_models = {
        "x3": {"parent_names" : ["x0"]}
    }
    sim.train(X, causal_graph)
    try:
        sim.run(changing_models=changing_models)
    except Exception:
        pass
    else:
        raise AssertionError


def test_cbsi_lingam_success(test_data, test_data_discrete):
    X, causal_graph, is_correct_co = test_data

    # constructor

    # X
    CBSILiNGAM(X, causal_graph)

    # X is pandas.DataFrame
    X_df = pd.DataFrame(X, columns=[f"x{i}" for i in range(X.shape[1])])
    CBSILiNGAM(X_df, causal_graph)

    # properties
    impl = CBSILiNGAM(X, causal_graph)
    assert is_correct_co(impl.causal_order_)
    assert impl.endog_names_ == [f"{i}" for i in range(len(causal_graph))]
    assert impl.discrete_endog_names_ == []
    assert impl.exog_length_ == len(X)

    # get_paraent_names
    parent = impl.get_parent_names("1")
    assert parent == ["0", "2"]
    parent = impl.get_parent_names("3")
    assert parent == []

    # get_data
    data = impl.get_data("0")
    assert np.all(data == X[:, [0]])

    # get_causal_order
    causal_order = impl.get_causal_order()
    assert is_correct_co(causal_order)


def test_cbsi_lingam_success2(test_data_discrete):
    X, causal_graph, is_correct_co, is_discrete = test_data_discrete

    # constructor

    # X
    CBSILiNGAM(X, causal_graph, is_discrete=is_discrete)

    # X is pandas.DataFrame
    X_df = pd.DataFrame(X, columns=[f"x{i}" for i in range(X.shape[1])])
    CBSILiNGAM(X_df, causal_graph, is_discrete)

    # properties

    impl = CBSILiNGAM(X, causal_graph, is_discrete=is_discrete)
    assert is_correct_co(impl.causal_order_)
    assert impl.endog_names_ == [f"{i}" for i in range(len(causal_graph))]
    assert len(impl.discrete_endog_names_) == 2

    # get_paraent_names
    parent = impl.get_parent_names("1")
    assert parent == ["0", "2"]
    parent = impl.get_parent_names("3")
    assert parent == []

    # get_data
    data = impl.get_data("0")
    assert np.all(data == X[:, [0]])

    # get_causal_order
    causal_order = impl.get_causal_order()
    assert is_correct_co(causal_order)


def test_cbsi_lingam_exception(test_data):
    X, causal_graph, _ = test_data

    # X: wrong dim
    try:
        CBSILiNGAM(X.reshape(1, *X.shape), causal_graph)
    except Exception:
        pass
    else:
        raise AssertionError

    # X: nan
    X_ = X.copy()
    X_[0, 0] = np.nan
    try:
        CBSILiNGAM(X_, causal_graph)
    except Exception:
        pass
    else:
        raise AssertionError

    # causal_graph: wrong shape
    try:
        CBSILiNGAM(X, np.concatenate([causal_graph] * 2))
    except Exception:
        pass
    else:
        raise AssertionError

    # causal_graph: wrong dim
    try:
        CBSILiNGAM(X, causal_graph.reshape(1, *causal_graph.shape))
    except Exception:
        pass
    else:
        raise AssertionError

    # causal_graph: cyclic
    causal_graph_cyclic = causal_graph.copy()
    causal_graph_cyclic[3, 5] = 1
    try:
        CBSILiNGAM(X, causal_graph_cyclic)
    except Exception:
        pass
    else:
        raise AssertionError

    # is_discrete: wrong length
    is_discrete = [False for i in range(len(causal_graph) + 1)]
    try:
        CBSILiNGAM(X, causal_graph, is_discrete=is_discrete)
    except Exception:
        pass
    else:
        raise AssertionError


def test_cbsi_unobs_success(test_data_unobs):
    X, causal_graph = test_data_unobs

    # X
    CBSIUnobsCommonCauseLiNGAM(X, causal_graph)

    # X is pandas.DataFrame
    X_df = pd.DataFrame(X, columns=[f"x{i}" for i in range(X.shape[1])])
    CBSIUnobsCommonCauseLiNGAM(X_df, causal_graph)


def test_cbsi_ts_success(test_data_ts):
    X, causal_graph, is_correct_co = test_data_ts

    # constructor

    # X
    CBSITimeSeriesLiNGAM(X, causal_graph)

    # X is pandas.DataFrame
    X_df = pd.DataFrame(X, columns=[f"x{i}" for i in range(X.shape[1])])
    CBSITimeSeriesLiNGAM(X_df, causal_graph)

    # properties

    impl = CBSITimeSeriesLiNGAM(X, causal_graph)
    assert impl.endog_names_ == [*[f"{i}[t]" for i in range(X.shape[1])], *[f"{i}[t-1]" for i in range(X.shape[1])]]

    # get_paraent_names
    parent = impl.get_parent_names("1[t]")
    assert parent == ["1[t-1]", "2[t-1]", "3[t-1]", "4[t-1]"]
    parent = impl.get_parent_names("1[t-1]")
    assert parent == []

    # get_data
    data = impl.get_data("1[t]")
    n_lags = len(causal_graph) - 1
    assert data.shape == (len(X) - n_lags, 1)

    # get_causal_order
    causal_order = impl.get_causal_order()
    assert is_correct_co(causal_order)


def test_cbsi_ts_exception(test_data_ts):
    X, causal_graph, is_correct_co = test_data_ts

    # X: wrong dim
    try:
        CBSITimeSeriesLiNGAM(X.reshape(1, *X.shape), causal_graph)
    except Exception:
        pass
    else:
        raise AssertionError

    # X: wrong n_features
    X_ = np.concatenate([X, X], axis=1)
    try:
        CBSITimeSeriesLiNGAM(X_, causal_graph)
    except Exception:
        pass
    else:
        raise AssertionError

    # causal_graph: wrong dim
    try:
        CBSITimeSeriesLiNGAM(X, causal_graph.reshape(1, *causal_graph.shape))
    except Exception:
        pass
    else:
        raise AssertionError

    # causal_graph: wrong length
    causal_graph_ = np.concatenate([causal_graph] * 2, axis=1)
    try:
        CBSITimeSeriesLiNGAM(X, causal_graph_)
    except Exception:
        pass
    else:
        raise AssertionError



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""

import numpy as np
from scipy.linalg import eigh
from scipy.sparse.linalg import eigsh
from sklearn.utils import check_array


def f_correlation(x, y):
    """Implementation of F-correlation [2]_

    References
    ----------
    .. [2] Bach, Francis R., and Michael I. Jordan. "Kernel independent component analysis."
        Journal of machine learning research 3.Jul (2002): 1-48.

    Parameters
    ----------
    x, y : array-like, shape (n_samples)
        Data, where ``n_samples`` is the number of samples.

    Returns
    -------
    float
        The valus of F-correlation.
    """

    x_ = check_array(x, ensure_2d=False)
    y_ = check_array(y, ensure_2d=False)
    n = x_.shape[0]

    if y_.shape[0] != n:
        raise ValueError("x and y must be the same size.")

    # Standardize
    x_ = (x_ - x_.mean()) / x_.std()
    y_ = (y_ - y_.mean()) / y_.std()

    X = [x_, y_]
    m = len(X)

    if n > 1000:
        kappa, sigma = [2e-3, 0.5]
    else:
        kappa, sigma = [2e-2, 1.0]

    Rs = []
    Us = []
    sizes = []
    for i in range(m):
        # Incomplete Cholesky decomposition
        G, P = _incomplete_cholesky(X[i], sigma, n * kappa * 1e-2)
        G = G[np.argsort(P), :]
        G = G - np.tile(np.mean(G, axis=0), (len(G), 1))

        # Singular value decomposition
        R, U = _svd(G, n, kappa)
        Rs.append(R)
        Us.append(U)
        sizes.append(R.shape[0])

    # make R_kappa
    R_kappa = np.eye(sum(sizes))
    st = np.cumsum(np.hstack([0, sizes]))
    st = st[:m]
    for i in range(1, m):
        for j in range(0, i):
            RUUR = np.diag(Rs[i]) @ (Us[i].T @ Us[j]) @ np.diag(Rs[j])
            R_kappa[st[i] : st[i] + sizes[i], st[j] : st[j] + sizes[j]] = RUUR
            R_kappa[st[j] : st[j] + sizes[j], st[i] : st[i] + sizes[i]] = RUUR.T

    # smallest eigenvalue of R_kappa
    w, _ = eigsh(R_kappa, 1, which="SM")

    return 1 - w[0]


def _squared_dist(x, y):
    """Squared euclidean distance matrix."""
    xx = x * x
    yy = y * y
    xy = x.reshape(-1, 1) * y
    xx = np.tile(xx.reshape(-1, 1), (1, len(yy)))
    yy = np.tile(yy, (len(xx), 1))
    return np.abs(xx + yy - 2 * xy)


def _incomplete_cholesky(x, sigma, tol):
    """Incomplete cholesky decomposition of the gram matrix with the gaussian kernel."""
    n = x.shape[0]
    i = 0
    Gjj = np.ones((n))
    G = np.empty((n, 0))
    P = np.array([i for i in range(n)])

    while np.sum(Gjj[i:n]) > tol:
        G = np.append(G, np.zeros((n, 1)), axis=1)
        if i > 0:
            jast = np.argmax(Gjj[i:n])
            jast = jast + i
            P[[i, jast]] = P[[jast, i]]
            G[[i, jast], :i] = G[[jast, i], :i]
        else:
            jast = 0

        G[i, i] = Gjj[jast]
        G[i, i] = np.sqrt(G[i, i])

        if i < n:
            dist = _squared_dist(x[P[(i + 1) : n]], x[[P[i]]])
            K = np.exp(-0.5 / sigma**2 * dist)
            if i > 0:
                sigG = G[(i + 1) : n, 0:i] @ G[i, 0:i].reshape(-1, 1)
                G[(i + 1) : n, [i]] = 1 / G[i, i] * (K - sigG)
            else:
                G[(i + 1) : n, [i]] = 1 / G[i, i] * K

        if i < n:
            Gjj[(i + 1) : n] = np.ones((n - (i + 1))) - np.sum(
                G[(i + 1) : n, 0 : i + 1] ** 2, axis=1
            )

        i = i + 1

    return G, P


def _svd(G, n, kappa):
    """Singular value decomposition."""
    eta = kappa * 1e-2
    D, A = eigh(G.T @ G)
    indexes = np.where(D >= n * eta)[0]
    order = np.argsort(D[indexes])
    order = order[::-1]
    indexes = indexes[order[0 : len(indexes)]]
    D = D[indexes]
    U = G @ (A[:, indexes] @ np.diag(np.sqrt(1.0 / (D))))
    R = D
    for j in range(len(D)):
        # regularized
        R[j] = D[j] / (n * kappa / 2 + D[j])
    return R, U



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""

import itertools
import numbers
import warnings

import numpy as np
from scipy.optimize import fmin_l_bfgs_b
from scipy.stats import pearsonr, shapiro, chi2
from sklearn.linear_model import LinearRegression
from sklearn.utils import check_array, resample

from .bootstrap import BootstrapResult
from .hsic import get_gram_matrix, get_kernel_width, hsic_test_gamma, hsic_teststat
from .utils import f_correlation, calculate_total_effect


class MultiGroupRCD:
    """Implementation of RCD Algorithm with multiple groups"""

    def __init__(
        self,
        max_explanatory_num=2,
        cor_alpha=0.01,
        ind_alpha=0.01,
        shapiro_alpha=0.01,
        MLHSICR=True,
        bw_method="mdbs",
        independence="hsic",
        ind_corr=0.5,
    ):
        """Construct a model.

        Parameters
        ----------
        max_explanatory_num : int, optional (default=2)
            Maximum number of explanatory variables.
        cor_alpha : float, optional (default=0.01)
            Alpha level for pearson correlation.
        ind_alpha : float, optional (default=0.01)
            Alpha level for HSIC.
        shapiro_alpha : float, optional (default=0.01)
            Alpha level for Shapiro-Wilk test.
        MLHSICR : bool, optional (default=True)
            If True, use MLHSICR for multiple regression, if False, use OLS for multiple regression.
        bw_method : str, optional (default=``mdbs``)
                The method used to calculate the bandwidth of the HSIC.

            * ``mdbs`` : Median distance between samples.
            * ``scott`` : Scott's Rule of Thumb.
            * ``silverman`` : Silverman's Rule of Thumb.
        independence : {'hsic', 'fcorr'}, optional (default='hsic')
            Methods to determine independence. If 'hsic' is set, test for independence by HSIC.
            If 'fcorr' is set, independence is determined by F-correlation.
        ind_corr : float, optional (default=0.5)
            The threshold value for determining independence by F-correlation;
            independence is determined when the value of F-correlation is below this threshold value.
        """
        # Check parameters
        if max_explanatory_num <= 0:
            raise ValueError("max_explanatory_num must be > 0.")

        if cor_alpha < 0:
            raise ValueError("cor_alpha must be >= 0.")

        if ind_alpha < 0:
            raise ValueError("ind_alpha must be >= 0.")

        if shapiro_alpha < 0:
            raise ValueError("shapiro_alpha must be >= 0.")

        if bw_method not in ("mdbs", "scott", "silverman"):
            raise ValueError("bw_method must be 'mdbs', 'scott' or 'silverman'.")

        if independence not in ("hsic", "fcorr"):
            raise ValueError("independence must be 'hsic' or 'fcorr'.")

        if ind_corr < 0.0:
            raise ValueError("ind_corr must be an float greater than 0.")

        self._max_explanatory_num = max_explanatory_num
        self._cor_alpha = cor_alpha
        self._ind_alpha = ind_alpha
        self._shapiro_alpha = shapiro_alpha
        self._MLHSICR = MLHSICR
        self._bw_method = bw_method
        self._ancestors_list = None
        self._adjacency_matrices = None
        self._independence = independence
        self._ind_corr = ind_corr

    def fit(self, X_list):
        """Fit the model to multiple datasets.

        Parameters
        ----------
        X_list : list, shape [X, ...]
            Multiple datasets for training, where ``X`` is an dataset.
            The shape of ''X'' is (n_samples, n_features),
            where ``n_samples`` is the number of samples and ``n_features`` is the number of features.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        # Check parameters
        X_list = self._check_X_list(X_list)

        # Extract a set of ancestors of each variable
        M = self._extract_ancestors(X_list)

        # Extract parents (direct causes) from the set of ancestors.
        P = self._extract_parents(X_list, M)

        # Find the pairs of variables affected by the same latent confounders.
        C = self._extract_vars_sharing_confounders(X_list, P)

        self._ancestors_list = M

        self._adjacency_matrices = []
        for X in X_list:
            adjacency_matrix = self._estimate_adjacency_matrix(X, P, C)
            self._adjacency_matrices.append(adjacency_matrix)

        return self

    def estimate_total_effect(self, X_list, from_index, to_index):
        """Estimate total effect using causal model.

        Parameters
        ----------
        X_list : array-like, shape (X, ...)
            Multiple datasets for training, where ``X`` is an dataset.
            The shape of ''X'' is (n_samples, n_features),
            where ``n_samples`` is the number of samples and ``n_features`` is the number of features.
        from_index :
            Index of source variable to estimate total effect.
        to_index :
            Index of destination variable to estimate total effect.

        Returns
        -------
        total_effect : float
            Estimated total effect.
        """
        # Check parameters
        X_list = self._check_X_list(X_list)

        # Check from/to ancestors
        if to_index in self._ancestors_list[from_index]:
            warnings.warn(
                f"The estimated causal effect may be incorrect because "
                f"the causal order of the destination variable (to_index={to_index}) "
                f"is earlier than the source variable (from_index={from_index})."
            )

        effects = []
        for X, am in zip(X_list, self._adjacency_matrices):
            # Check confounders
            if True in np.isnan(am[from_index]):
                warnings.warn(
                    f"The estimated causal effect may be incorrect because "
                    f"the source variable (from_index={from_index}) is influenced by confounders."
                )
                return effects

            # from_index + parents indices
            parents = np.where(np.abs(am[from_index]) > 0)[0]
            predictors = [from_index]
            predictors.extend(parents)

            # Estimate total effect
            lr = LinearRegression()
            lr.fit(X[:, predictors], X[:, to_index])

            effects.append(lr.coef_[0])

        return effects

    def estimate_total_effect2(self, from_index, to_index):
        """Estimate total effect using causal model.

        Parameters
        ----------
        from_index :
            Index of source variable to estimate total effect.
        to_index :
            Index of destination variable to estimate total effect.

        Returns
        -------
        total_effect : float
            Estimated total effect.
        """
        # Check from/to ancestors
        if to_index in self._ancestors_list[from_index]:
            warnings.warn(
                f"The estimated causal effect may be incorrect because "
                f"the causal order of the destination variable (to_index={to_index}) "
                f"is earlier than the source variable (from_index={from_index})."
            )

        effects = []
        for am in self._adjacency_matrices:
            # Check confounders
            if True in np.isnan(am[from_index]):
                warnings.warn(
                    f"The estimated causal effect may be incorrect because "
                    f"the source variable (from_index={from_index}) is influenced by confounders."
                )
                return effects

            effect = calculate_total_effect(am, from_index, to_index)
            effects.append(effect)

        return effects

    def get_error_independence_p_values(self, X_list):
        """Calculate the p-value matrix of independence between error variables.

        Parameters
        ----------
        X_list : array-like, shape (X, ...)
            Multiple datasets for training, where ``X`` is an dataset.
            The shape of ''X'' is (n_samples, n_features),
            where ``n_samples`` is the number of samples and ``n_features`` is the number of features.

        Returns
        -------
        independence_p_values : array-like, shape (n_datasets, n_features, n_features)
            p-value matrix of independence between error variables.
        """
        # Check parameters
        X_list = self._check_X_list(X_list)

        p_values = np.zeros([len(X_list), self._n_features, self._n_features])
        for d, (X, am) in enumerate(zip(X_list, self._adjacency_matrices)):
            n_samples = X.shape[0]
            E = X - np.dot(am, X.T).T
            nan_cols = list(set(np.argwhere(np.isnan(am)).ravel()))
            for i, j in itertools.combinations(range(self._n_features), 2):
                if i in nan_cols or j in nan_cols:
                    p_values[d, i, j] = np.nan
                    p_values[d, j, i] = np.nan
                else:
                    _, p_value = hsic_test_gamma(
                        np.reshape(E[:, i], [n_samples, 1]),
                        np.reshape(E[:, j], [n_samples, 1]),
                    )
                    p_values[d, i, j] = p_value
                    p_values[d, j, i] = p_value

        return p_values

    def bootstrap(self, X_list, n_sampling):
        """Evaluate the statistical reliability of DAG based on the bootstrapping.

        Parameters
        ----------
        X_list : array-like, shape (X, ...)
            Multiple datasets for training, where ``X`` is an dataset.
            The shape of ''X'' is (n_samples, n_features),
            where ``n_samples`` is the number of samples and ``n_features`` is the number of features.
        n_sampling : int
            Number of bootstrapping samples.

        Returns
        -------
        results : array-like, shape (BootstrapResult, ...)
            Returns the results of bootstrapping for multiple datasets.
        """
        # Check parameters
        X_list = self._check_X_list(X_list)

        if isinstance(n_sampling, (numbers.Integral, np.integer)):
            if not 0 < n_sampling:
                raise ValueError("n_sampling must be an integer greater than 0.")
        else:
            raise ValueError("n_sampling must be an integer greater than 0.")

        # Bootstrapping
        adjacency_matrices_list = np.zeros(
            [len(X_list), n_sampling, self._n_features, self._n_features]
        )
        total_effects_list = np.zeros(
            [len(X_list), n_sampling, self._n_features, self._n_features]
        )
        for n in range(n_sampling):
            resampled_X_list = [resample(X) for X in X_list]
            self.fit(resampled_X_list)

            for i, am in enumerate(self._adjacency_matrices):
                adjacency_matrices_list[i][n] = am

            # Calculate total effects
            for to, ancestors in enumerate(self._ancestors_list):
                for from_ in ancestors:
                    effects = self.estimate_total_effect2(from_, to)
                    for i, effect in enumerate(effects):
                        total_effects_list[i, n, to, from_] = effect

        result_list = []
        for am, te in zip(adjacency_matrices_list, total_effects_list):
            result_list.append(BootstrapResult(am, te))

        return result_list

    def _check_X_list(self, X_list):
        """Check input X list."""
        if not isinstance(X_list, list):
            raise ValueError("X_list must be a list.")

        if len(X_list) < 2:
            raise ValueError("X_list must be a list containing at least two items")

        self._k = len(X_list)
        self._n_features = check_array(X_list[0]).shape[1]
        X_list_ = []
        for X in X_list:
            X_ = check_array(X)
            if X_.shape[1] != self._n_features:
                raise ValueError(
                    "X_list must be a list with the same number of features"
                )
            X_list_.append(X_)

        return X_list_

    def _get_common_ancestors(self, M, U):
        """Get the set of common ancestors of U"""
        Mj_list = [M[xj] for xj in U]
        return set.intersection(*Mj_list)

    def _get_resid_and_coef(self, X, endog_idx, exog_idcs):
        """Get the residuals and coefficients of the ordinary least squares method"""
        lr = LinearRegression()
        lr.fit(X[:, exog_idcs], X[:, endog_idx])
        resid = X[:, endog_idx] - lr.predict(X[:, exog_idcs])
        return resid, lr.coef_

    def _get_residual_matrix(self, X, U, H_U):
        if len(H_U) == 0:
            return X

        Y = np.zeros_like(X)
        for xj in U:
            Y[:, xj], _ = self._get_resid_and_coef(X, xj, list(H_U))
        return Y

    def _is_non_gaussianity(self, Y_list, U):
        """Test whether a variable is generated from a non-Gaussian process using the Shapiro-Wilk test"""
        for xj in U:
            fisher_stat = 0
            for Y in Y_list:
                shapiro_p = shapiro(Y[:, xj])[1]
                fisher_stat += np.inf if shapiro_p == 0 else -2 * np.log(shapiro_p)

            fisher_p = chi2.sf(fisher_stat, df=2 * self._k)

            if fisher_p > self._shapiro_alpha:
                return False

        return True

    def _is_correlated(self, Y_list, U):
        # Estimate that the two variables are linearly correlated using the Pearson's correlation
        for xi, xj in itertools.combinations(U, 2):
            fisher_stat = 0
            for Y in Y_list:
                corr_p = pearsonr(Y[:, xi], Y[:, xj])[1]
                fisher_stat += np.inf if corr_p == 0 else -2 * np.log(corr_p)

            fisher_p = chi2.sf(fisher_stat, df=2 * self._k)

            if fisher_p >= self._cor_alpha:
                return False
        return True

    def _exists_ancestor_in_U(self, M, U, xi, xj_list):
        # Check xi is not in Mj, the ancestor of xj.
        for xj in xj_list:
            if xi in M[xj]:
                return True

        # Check if xj_list is a subset of Mi, the ancestor of xi.
        if set(xj_list) == set(xj_list) & M[xi]:
            return True
        return False

    def _is_independent(self, X_list, Y_list, xj):
        if self._independence == "hsic":
            fisher_stat = 0
            for i, Y in enumerate(Y_list):
                _, hsic_p = hsic_test_gamma(
                    X_list[i], Y[:, xj], bw_method=self._bw_method
                )
                fisher_stat += np.inf if hsic_p == 0 else -2 * np.log(hsic_p)

            fisher_p = chi2.sf(fisher_stat, df=2 * self._k)

            is_independent = fisher_p > self._ind_alpha

        elif self._independence == "fcorr":
            max_f_corr = 0.0
            for i, Y in enumerate(Y_list):
                f_corr = f_correlation(X_list[i], Y[:, xj])
                if f_corr > max_f_corr:
                    max_f_corr = f_corr

            is_independent = f_corr < self._ind_corr

        return is_independent

    def _is_independent_of_resid(self, Y_list, xi, xj_list):
        """Check whether the residuals obtained from multiple regressions are independent"""
        # Multiple Regression with OLS.
        resid_list = []
        for Y in Y_list:
            resid, _ = self._get_resid_and_coef(Y, xi, xj_list)
            resid_list.append(resid)

        is_all_independent = True
        for xj in xj_list:
            is_independent = self._is_independent(resid_list, Y_list, xj)
            if not is_independent:
                is_all_independent = False
                break

        if is_all_independent:
            return True
        elif len(xj_list) == 1 or self._MLHSICR is False:
            return False

        # Multiple Regression with MLHSICR.
        resid_list = []
        for Y in Y_list:
            resid, _ = self._get_resid_and_coef_by_MLHSICR(Y, xi, xj_list)
            resid_list.append(resid)

        for xj in xj_list:
            is_independent = self._is_independent(resid_list, Y_list, xj)
            if not is_independent:
                return False

        return True

    def _get_resid_and_coef_by_MLHSICR(self, Y, xi, xj_list):
        """Get the residuals and coefficients by minimizing the sum of HSICs using the L-BFGS method."""
        n_samples = Y.shape[0]
        width_list = []
        Lc_list = []
        for xj in xj_list:
            yj = np.reshape(Y[:, xj], [n_samples, 1])
            width_xj = get_kernel_width(yj)
            _, Lc = get_gram_matrix(yj, width_xj)

            width_list.append(width_xj)
            Lc_list.append(Lc)

        _, initial_coef = self._get_resid_and_coef(Y, xi, xj_list)
        width_xi = get_kernel_width(np.reshape(Y[:, xi], [n_samples, 1]))

        # Calculate the sum of the Hilbert-Schmidt independence criterion
        def sum_empirical_hsic(coef):
            resid = Y[:, xi]
            width = width_xi
            for j, xj in enumerate(xj_list):
                resid = resid - coef[j] * Y[:, xj]
                width = width - coef[j] * width_list[j]
            _, Kc = get_gram_matrix(np.reshape(resid, [n_samples, 1]), width)

            objective = 0.0
            for j, xj in enumerate(xj_list):
                objective += hsic_teststat(Kc, Lc_list[j], n_samples)
            return objective

        # Estimate coefficients by minimizing the sum of HSICs using the L-BFGS method.
        coefs, _, _ = fmin_l_bfgs_b(
            func=sum_empirical_hsic, x0=initial_coef, approx_grad=True
        )

        resid = Y[:, xi]
        for j, xj in enumerate(xj_list):
            resid = resid - coefs[j] * Y[:, xj]
        return resid, coefs

    def _extract_ancestors(self, X_list):
        """Extract a set of ancestors of each variable"""
        M = [set() for i in range(self._n_features)]
        l = 1
        hu_history = {}

        while True:
            changed = False
            U_list = itertools.combinations(range(self._n_features), l + 1)
            for U in U_list:
                U = list(U)
                U.sort()

                # Get the set of common ancestors of U
                H_U = self._get_common_ancestors(M, U)

                if tuple(U) in hu_history and H_U == hu_history[tuple(U)]:
                    continue

                # Y_list = np.zeros(X_list.shape)
                Y_list = []
                for X in X_list:
                    Y_list.append(self._get_residual_matrix(X, U, H_U))

                # Test whether a variable is generated from a non-Gaussian process using the Shapiro-Wilk test
                if not self._is_non_gaussianity(Y_list, U):
                    continue

                # Estimate that the two variables are linearly correlated using the Pearson's correlation
                if not self._is_correlated(Y_list, U):
                    continue

                sink_set = []
                for xi in U:
                    xj_list = list(set(U) - set([xi]))
                    if self._exists_ancestor_in_U(M, U, xi, xj_list):
                        continue

                    # Check whether the residuals obtained from multiple regressions are independent
                    if self._is_independent_of_resid(Y_list, xi, xj_list):
                        sink_set.append(xi)

                if len(sink_set) == 1:
                    xi = sink_set[0]
                    xj_list = list(set(U) - set(sink_set))

                    if not M[xi] == M[xi] | set(xj_list):
                        M[xi] = M[xi] | set(xj_list)
                        changed = True

                hu_history[tuple(U)] = H_U

            if changed:
                l = 1
            elif l < self._max_explanatory_num:
                l += 1
            else:
                break

        return M

    def _extract_parents(self, X_list, M):
        """Extract parents (direct causes) from a set of ancestors"""
        P = [set() for i in range(self._n_features)]

        for xi in range(self._n_features):
            for xj in M[xi]:
                # Check if xj is the parent of xi
                if self._is_parent(X_list, M, xj, xi):
                    P[xi].add(xj)

        return P

    def _is_parent(self, X_list, M, xj, xi):
        # Check if zi and wj are correlated
        fisher_stat = 0
        for X in X_list:
            if len(M[xi] - set([xj])) > 0:
                zi, _ = self._get_resid_and_coef(X, xi, list(M[xi] - set([xj])))
            else:
                zi = X[:, xi]

            if len(M[xi] & M[xj]) > 0:
                wj, _ = self._get_resid_and_coef(X, xj, list(M[xi] & M[xj]))
            else:
                wj = X[:, xj]

            corr_p = pearsonr(wj, zi)[1]
            fisher_stat += np.inf if corr_p == 0 else -2 * np.log(corr_p)

        fisher_p = chi2.sf(fisher_stat, df=2 * self._k)

        return fisher_p < self._cor_alpha

    def _get_resid_to_parent(self, X, idx, P):
        if len(P[idx]) == 0:
            return X[:, idx]

        resid, _ = self._get_resid_and_coef(X, idx, list(P[idx]))
        return resid

    def _extract_vars_sharing_confounders(self, X_list, P):
        """Find the pairs of variables affected by the same latent confounders."""
        C = [set() for i in range(self._n_features)]

        for i, j in itertools.combinations(range(self._n_features), 2):
            if (i in P[j]) or (j in P[i]):
                continue

            fisher_stat = 0
            for X in X_list:
                resid_xi = self._get_resid_to_parent(X, i, P)
                resid_xj = self._get_resid_to_parent(X, j, P)
                corr_p = pearsonr(resid_xi, resid_xj)[1]
                fisher_stat += np.inf if corr_p == 0 else -2 * np.log(corr_p)

            fisher_p = chi2.sf(fisher_stat, df=2 * self._k)

            if fisher_p < self._cor_alpha:
                C[i].add(j)
                C[j].add(i)

        return C

    def _estimate_adjacency_matrix(self, X, P, C):
        # Check parents
        B = np.zeros([self._n_features, self._n_features], dtype="float64")
        for xi in range(self._n_features):
            xj_list = list(P[xi])
            xj_list.sort()
            if len(xj_list) == 0:
                continue

            _, coef = self._get_resid_and_coef(X, xi, xj_list)
            for j, xj in enumerate(xj_list):
                B[xi, xj] = coef[j]

        # Check confounders
        for xi in range(self._n_features):
            xj_list = list(C[xi])
            xj_list.sort()
            if len(xj_list) == 0:
                continue

            for xj in xj_list:
                B[xi, xj] = np.nan

        return B

    @property
    def adjacency_matrices_(self):
        """Estimated adjacency matrices.

        Returns
        -------
        adjacency_matrices_ : array-like, shape (B, ...)
            The list of adjacency matrix B for multiple datasets.
            The shape of B is (n_features, n_features), where
            n_features is the number of features.
        """
        return self._adjacency_matrices

    @property
    def ancestors_list_(self):
        """Estimated ancestors list.

        Returns
        -------
        ancestors_list_ : array-like, shape (n_features)
            The list of causal ancestors sets, where
            n_features is the number of features.
        """
        return self._ancestors_list



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""

import numpy as np
from sklearn.utils import check_array

from .base import _BaseLiNGAM
from .hsic import hsic_test_gamma


class RESIT(_BaseLiNGAM):
    """Implementation of RESIT(regression with subsequent independence test) Algorithm [1]_

    References
    ----------
    .. [1] Jonas Peters, Joris M Mooij, Dominik Janzing, and Bernhard Sch ̈olkopf.
       Causal discovery with continuous additive noise models.
       Journal of Machine Learning Research, 15:2009-2053, 2014.

    Notes
    -----
    RESIT algorithm returns an **adjacency matrix consisting of zeros or ones**,
    rather than an adjacency matrix consisting of causal coefficients,
    in order to estimate nonlinear causality.
    """

    def __init__(self, regressor, random_state=None, prior_knowledge=None, alpha=0.01):
        """Construct a RESIT model.

        Parameters
        ----------
        regressor : regressor object implementing 'fit' and 'predict' function (default=None)
            Regressor to compute residuals.
            This regressor object must have ``fit`` method and ``predict`` function like scikit-learn's model.
        random_state : int, optional (default=None)
            ``random_state`` is the seed used by the random number generator.
        prior_knowledge : array-like, shape (n_features, n_features), optional (default=None)
            Prior knowledge used for causal discovery, where ``n_features`` is the number of features.

            The elements of prior knowledge matrix are defined as follows [1]_:

            * ``0`` : :math:`x_i` does not have a directed path to :math:`x_j`
            * ``1`` : :math:`x_i` has a directed path to :math:`x_j`
            * ``-1`` : No prior knowledge is available to know if either of the two cases above (0 or 1) is true.
        alpha : float, optional (default=0.01)
            Alpha level for HSIC independence test when removing superfluous edges.
        """
        # Check parameters
        if regressor is None:
            raise ValueError("Specify regression model in 'regressor'.")
        else:
            if not (hasattr(regressor, "fit") and hasattr(regressor, "predict")):
                raise ValueError("'regressor' has no fit or predict method.")

        if alpha < 0.0:
            raise ValueError("alpha must be an float greater than 0.")

        super().__init__(random_state)
        self._Aknw = prior_knowledge
        self._alpha = alpha
        self._reg = regressor

        if self._Aknw is not None:
            self._Aknw = check_array(self._Aknw)
            self._Aknw = np.where(self._Aknw < 0, np.nan, self._Aknw)

    def fit(self, X):
        """Fit the model to X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        # Check parameters
        X = check_array(X)
        n_features = X.shape[1]

        # Check prior knowledge
        if self._Aknw is not None:
            if (n_features, n_features) != self._Aknw.shape:
                raise ValueError(
                    "The shape of prior knowledge must be (n_features, n_features)"
                )
            else:
                # Extract all partial orders in prior knowledge matrix
                self._partial_orders = self._extract_partial_orders(self._Aknw)

        # Determine topological order
        pa, pi = self._estimate_order(X)

        # Remove superfluous edges
        pa = self._remove_edges(X, pa, pi)

        # Create adjacency matrix from parent-child relationship
        adjacency_matrix = np.zeros([n_features, n_features])
        for i, parents in pa.items():
            for p in parents:
                adjacency_matrix[i, p] = 1

        self._causal_order = pi
        self._adjacency_matrix = adjacency_matrix
        return self

    def _estimate_order(self, X):
        """Determine topological order"""
        S = np.arange(X.shape[1])
        pa = {}
        pi = []

        for _ in range(X.shape[1]):
            Sc = self._search_candidate(S)

            if len(Sc)==1:
                k = Sc[0]
            else:
                hsic_stats = []
                for k in Sc:
                    # Regress Xk on {Xi}
                    predictors = [i for i in S if i != k]
                    self._reg.fit(X[:, predictors], X[:, k])
                    residual = X[:, k] - self._reg.predict(X[:, predictors])

                    # Measure dependence between residuals and {Xi}
                    hsic_stat, _ = hsic_test_gamma(residual, X[:, predictors])
                    hsic_stats.append(hsic_stat)

                k = Sc[np.argmin(hsic_stats)]

            S = S[S != k]
            pa[k] = S.tolist()
            pi.insert(0, k)

            # Update partial orders
            if self._Aknw is not None:
                self._partial_orders = self._partial_orders[self._partial_orders[:, 1] != k]

        # The topological order after exploration is revised based on prior knowledge.
        if self._Aknw is not None:
            _pk = self._Aknw.copy()
            np.fill_diagonal(_pk, 0)

            pk_pa = {}
            for i, parents in pa.items():
                new_parents = [p for p in parents if _pk[i, p] != 0]
                pk_pa[i] = new_parents
            pa = pk_pa

        return pa, pi

    def _remove_edges(self, X, pa, pi):
        """Remove superfluous edges"""
        for k in range(1, X.shape[1]):
            parents = pa[pi[k]].copy()

            for l in parents:
                # Regress Xk on {Xi}
                predictors = [i for i in pa[pi[k]] if i != l]
                if len(predictors) >= 1:
                    self._reg.fit(X[:, predictors], X[:, pi[k]])
                    residual = X[:, pi[k]] - self._reg.predict(X[:, predictors])
                else:
                    residual = X[:, pi[k]]
                # Measure dependence between residuals and {Xi}
                _, hsic_p = hsic_test_gamma(residual, X[:, parents])
                # eliminate edge
                if hsic_p > self._alpha:
                    pa[pi[k]].remove(l)

        return pa

    def _extract_partial_orders(self, pk):
        """Extract partial orders from prior knowledge."""
        path_pairs = np.array(np.where(pk == 1)).transpose()
        no_path_pairs = np.array(np.where(pk == 0)).transpose()

        # Check for inconsistencies in pairs with path
        check_pairs = np.concatenate([path_pairs, path_pairs[:, [1, 0]]])
        if len(check_pairs) > 0:
            pairs, counts = np.unique(check_pairs, axis=0, return_counts=True)
            if len(pairs[counts > 1]) > 0:
                raise ValueError(
                    f"The prior knowledge contains inconsistencies at the following indices: {pairs[counts>1].tolist()}"
                )

        # Check for inconsistencies in pairs without path
        # If there are duplicate pairs without path, they cancel out and are not ordered.
        check_pairs = np.concatenate([no_path_pairs, no_path_pairs[:, [1, 0]]])
        if len(check_pairs) > 0:
            pairs, counts = np.unique(check_pairs, axis=0, return_counts=True)
            check_pairs = np.concatenate([no_path_pairs, pairs[counts > 1]])
            pairs, counts = np.unique(check_pairs, axis=0, return_counts=True)
            no_path_pairs = pairs[counts < 2]

        check_pairs = np.concatenate([path_pairs, no_path_pairs[:, [1, 0]]])
        if len(check_pairs) == 0:
            # If no pairs are extracted from the specified prior knowledge,
            return check_pairs

        pairs = np.unique(check_pairs, axis=0)
        return pairs[:, [1, 0]]  # [to, from] -> [from, to]

    def _search_candidate(self, S):
        """Search for candidate features"""
        # If no prior knowledge is specified, nothing to do.
        if self._Aknw is None:
            return S

        # Candidate features that are not to the left of the partial orders
        if len(self._partial_orders) != 0:
            Sc = [i for i in S if i not in self._partial_orders[:, 0]]
            return Sc

        return S

    def estimate_total_effect(self, X, from_index, to_index):
        """Estimate total effect using causal model.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Original data, where n_samples is the number of samples
            and n_features is the number of features.
        from_index :
            Index of source variable to estimate total effect.
        to_index :
            Index of destination variable to estimate total effect.

        Returns
        -------
        total_effect : float
            **Because RESIT is a nonlinear algorithm, it cannot estimate the total effect and always returns a value of zero**
        """
        return 0

    def get_error_independence_p_values(self, X):
        """Calculate the p-value matrix of independence between error variables.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Original data, where n_samples is the number of samples
            and n_features is the number of features.

        Returns
        -------
        independence_p_values : array-like, shape (n_features, n_features)
            **RESIT always returns zero**
        """
        n_features = X.shape[1]
        p_values = np.zeros([n_features, n_features])
        return p_values



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""

import itertools
import numbers
import warnings

import numpy as np
from scipy.optimize import fmin_l_bfgs_b
from scipy.stats import pearsonr, shapiro
from sklearn.linear_model import LinearRegression
from sklearn.utils import check_array, resample

from .bootstrap import BootstrapResult
from .hsic import get_gram_matrix, get_kernel_width, hsic_test_gamma, hsic_teststat
from .utils import f_correlation, calculate_total_effect


class RCD:
    """Implementation of RCD Algorithm [1]_

    References
    ----------
    .. [1] T.N.Maeda and S.Shimizu. RCD: Repetitive causal discovery of
       linear non-Gaussian acyclic models with latent confounders.
       In Proc. 23rd International Conference on Artificial Intelligence and
       Statistics (AISTATS2020), Palermo, Sicily, Italy. PMLR  108:735-745, 2020.
    """

    def __init__(
        self,
        max_explanatory_num=2,
        cor_alpha=0.01,
        ind_alpha=0.01,
        shapiro_alpha=0.01,
        MLHSICR=False,
        bw_method="mdbs",
        independence="hsic",
        ind_corr=0.5,
    ):
        """Construct a RCD model.

        Parameters
        ----------
        max_explanatory_num : int, optional (default=2)
            Maximum number of explanatory variables.
        cor_alpha : float, optional (default=0.01)
            Alpha level for pearson correlation.
        ind_alpha : float, optional (default=0.01)
            Alpha level for HSIC.
        shapiro_alpha : float, optional (default=0.01)
            Alpha level for Shapiro-Wilk test.
        MLHSICR : bool, optional (default=False)
            If True, use MLHSICR for multiple regression, if False, use OLS for multiple regression.
        bw_method : str, optional (default=``mdbs``)
            The method used to calculate the bandwidth of the HSIC.

            * ``mdbs`` : Median distance between samples.
            * ``scott`` : Scott's Rule of Thumb.
            * ``silverman`` : Silverman's Rule of Thumb.
        independence : {'hsic', 'fcorr'}, optional (default='hsic')
            Methods to determine independence. If 'hsic' is set, test for independence by HSIC.
            If 'fcorr' is set, independence is determined by F-correlation.
        ind_corr : float, optional (default=0.5)
            The threshold value for determining independence by F-correlation;
            independence is determined when the value of F-correlation is below this threshold value.
        """
        # Check parameters
        if max_explanatory_num <= 0:
            raise ValueError("max_explanatory_num must be > 0.")

        if cor_alpha < 0:
            raise ValueError("cor_alpha must be >= 0.")

        if ind_alpha < 0:
            raise ValueError("ind_alpha must be >= 0.")

        if shapiro_alpha < 0:
            raise ValueError("shapiro_alpha must be >= 0.")

        if bw_method not in ("mdbs", "scott", "silverman"):
            raise ValueError("bw_method must be 'mdbs', 'scott' or 'silverman'.")

        if independence not in ("hsic", "fcorr"):
            raise ValueError("independence must be 'hsic' or 'fcorr'.")

        if ind_corr < 0.0:
            raise ValueError("ind_corr must be an float greater than 0.")

        self._max_explanatory_num = max_explanatory_num
        self._cor_alpha = cor_alpha
        self._ind_alpha = ind_alpha
        self._shapiro_alpha = shapiro_alpha
        self._MLHSICR = MLHSICR
        self._bw_method = bw_method
        self._ancestors_list = None
        self._adjacency_matrix = None
        # DEBAG: self._adjacency_matrix_no_lc = None
        self._independence = independence
        self._ind_corr = ind_corr

    def fit(self, X):
        """Fit the model to X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        # Check parameters
        X = check_array(X)

        # Extract a set of ancestors of each variable
        M = self._extract_ancestors(X)

        # Extract parents (direct causes) from the set of ancestors.
        P = self._extract_parents(X, M)

        # Find the pairs of variables affected by the same latent confounders.
        C = self._extract_vars_sharing_confounders(X, P)

        self._ancestors_list = M
        return self._estimate_adjacency_matrix(X, P, C)

    def _get_common_ancestors(self, M, U):
        """Get the set of common ancestors of U"""
        Mj_list = [M[xj] for xj in U]
        return set.intersection(*Mj_list)

    def _get_resid_and_coef(self, X, endog_idx, exog_idcs):
        """Get the residuals and coefficients of the ordinary least squares method"""
        lr = LinearRegression()
        lr.fit(X[:, exog_idcs], X[:, endog_idx])
        resid = X[:, endog_idx] - lr.predict(X[:, exog_idcs])
        return resid, lr.coef_

    def _get_residual_matrix(self, X, U, H_U):
        if len(H_U) == 0:
            return X

        Y = np.zeros_like(X)
        for xj in U:
            Y[:, xj], _ = self._get_resid_and_coef(X, xj, list(H_U))
        return Y

    def _is_non_gaussianity(self, Y, U):
        """Test whether a variable is generated from a non-Gaussian process using the Shapiro-Wilk test"""
        for xj in U:
            if shapiro(Y[:, xj])[1] > self._shapiro_alpha:
                return False
        return True

    def _is_correlated(self, a, b):
        """Estimate that the two variables are linearly correlated using the Pearson's correlation"""
        return pearsonr(a, b)[1] < self._cor_alpha

    def _exists_ancestor_in_U(self, M, U, xi, xj_list):
        # Check xi is not in Mj, the ancestor of xj.
        for xj in xj_list:
            if xi in M[xj]:
                return True

        # Check if xj_list is a subset of Mi, the ancestor of xi.
        if set(xj_list) == set(xj_list) & M[xi]:
            return True
        return False

    def _is_independent(self, X, Y):
        is_independent = False
        if self._independence == "hsic":
            _, p = hsic_test_gamma(X, Y, bw_method=self._bw_method)
            is_independent = p > self._ind_alpha
        elif self._independence == "fcorr":
            f_corr = f_correlation(X, Y)
            is_independent = f_corr < self._ind_corr
        return is_independent

    def _get_resid_and_coef_by_MLHSICR(self, Y, xi, xj_list):
        """Get the residuals and coefficients by minimizing the sum of HSICs using the L-BFGS method."""
        n_samples = Y.shape[0]
        width_list = []
        Lc_list = []
        for xj in xj_list:
            yj = np.reshape(Y[:, xj], [n_samples, 1])
            width_xj = get_kernel_width(yj)
            _, Lc = get_gram_matrix(yj, width_xj)

            width_list.append(width_xj)
            Lc_list.append(Lc)

        _, initial_coef = self._get_resid_and_coef(Y, xi, xj_list)
        width_xi = get_kernel_width(np.reshape(Y[:, xi], [n_samples, 1]))

        # Calculate the sum of the Hilbert-Schmidt independence criterion
        def sum_empirical_hsic(coef):
            resid = Y[:, xi]
            width = width_xi
            for j, xj in enumerate(xj_list):
                resid = resid - coef[j] * Y[:, xj]
                width = width - coef[j] * width_list[j]
            _, Kc = get_gram_matrix(np.reshape(resid, [n_samples, 1]), width)

            objective = 0.0
            for j, xj in enumerate(xj_list):
                objective += hsic_teststat(Kc, Lc_list[j], n_samples)
            return objective

        # Estimate coefficients by minimizing the sum of HSICs using the L-BFGS method.
        coefs, _, _ = fmin_l_bfgs_b(
            func=sum_empirical_hsic, x0=initial_coef, approx_grad=True
        )

        resid = Y[:, xi]
        for j, xj in enumerate(xj_list):
            resid = resid - coefs[j] * Y[:, xj]
        return resid, coefs

    def _is_independent_of_resid(self, Y, xi, xj_list):
        """Check whether the residuals obtained from multiple regressions are independent"""
        n_samples = Y.shape[0]

        # Multiple Regression with OLS.
        is_all_independent = True
        resid, _ = self._get_resid_and_coef(Y, xi, xj_list)
        for xj in xj_list:
            if not self._is_independent(
                np.reshape(resid, [n_samples, 1]), np.reshape(Y[:, xj], [n_samples, 1])
            ):
                is_all_independent = False
                break

        if is_all_independent:
            return True
        elif len(xj_list) == 1 or self._MLHSICR is False:
            return False

        # Multiple Regression with MLHSICR.
        resid, _ = self._get_resid_and_coef_by_MLHSICR(Y, xi, xj_list)
        for xj in xj_list:
            if not self._is_independent(
                np.reshape(resid, [n_samples, 1]), np.reshape(Y[:, xj], [n_samples, 1])
            ):
                return False
        return True

    def _extract_ancestors(self, X):
        """Extract a set of ancestors of each variable"""
        n_features = X.shape[1]
        M = [set() for i in range(n_features)]
        l = 1
        hu_history = {}

        while True:
            changed = False
            U_list = itertools.combinations(range(n_features), l + 1)
            for U in U_list:
                U = list(U)
                U.sort()

                # Get the set of common ancestors of U
                H_U = self._get_common_ancestors(M, U)

                if tuple(U) in hu_history and H_U == hu_history[tuple(U)]:
                    continue

                Y = self._get_residual_matrix(X, U, H_U)

                # Test whether a variable is generated from a non-Gaussian process using the Shapiro-Wilk test
                if not self._is_non_gaussianity(Y, U):
                    continue

                # Estimate that the two variables are linearly correlated using the Pearson's correlation
                is_cor = True
                for xi, xj in itertools.combinations(U, 2):
                    if not self._is_correlated(Y[:, xi], Y[:, xj]):
                        is_cor = False
                        break
                if not is_cor:
                    continue

                sink_set = []
                for xi in U:
                    xj_list = list(set(U) - set([xi]))
                    if self._exists_ancestor_in_U(M, U, xi, xj_list):
                        continue

                    # Check whether the residuals obtained from multiple regressions are independent
                    if self._is_independent_of_resid(Y, xi, xj_list):
                        sink_set.append(xi)

                if len(sink_set) == 1:
                    xi = sink_set[0]
                    xj_list = list(set(U) - set(sink_set))

                    if not M[xi] == M[xi] | set(xj_list):
                        M[xi] = M[xi] | set(xj_list)
                        changed = True

                hu_history[tuple(U)] = H_U

            if changed:
                l = 1
            elif l < self._max_explanatory_num:
                l += 1
            else:
                break

        return M

    def _is_parent(self, X, M, xj, xi):
        if len(M[xi] - set([xj])) > 0:
            zi, _ = self._get_resid_and_coef(X, xi, list(M[xi] - set([xj])))
        else:
            zi = X[:, xi]

        if len(M[xi] & M[xj]) > 0:
            wj, _ = self._get_resid_and_coef(X, xj, list(M[xi] & M[xj]))
        else:
            wj = X[:, xj]

        # Check if zi and wj are correlated
        return self._is_correlated(wj, zi)

    def _extract_parents(self, X, M):
        """Extract parents (direct causes) from a set of ancestors"""
        n_features = X.shape[1]
        P = [set() for i in range(n_features)]

        for xi in range(n_features):
            for xj in M[xi]:
                # Check if xj is the parent of xi
                if self._is_parent(X, M, xj, xi):
                    P[xi].add(xj)

        return P

    def _get_resid_to_parent(self, X, idx, P):
        if len(P[idx]) == 0:
            return X[:, idx]

        resid, _ = self._get_resid_and_coef(X, idx, list(P[idx]))
        return resid

    def _extract_vars_sharing_confounders(self, X, P):
        """Find the pairs of variables affected by the same latent confounders."""
        n_features = X.shape[1]
        C = [set() for i in range(n_features)]

        for i, j in itertools.combinations(range(n_features), 2):
            if (i in P[j]) or (j in P[i]):
                continue
            resid_xi = self._get_resid_to_parent(X, i, P)
            resid_xj = self._get_resid_to_parent(X, j, P)
            if self._is_correlated(resid_xi, resid_xj):
                C[i].add(j)
                C[j].add(i)

        return C

    def _estimate_adjacency_matrix(self, X, P, C):
        """Estimate adjacency matrix by causal parents and confounders.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        # Check parents
        n_features = X.shape[1]
        B = np.zeros([n_features, n_features], dtype="float64")
        for xi in range(n_features):
            xj_list = list(P[xi])
            xj_list.sort()
            if len(xj_list) == 0:
                continue

            _, coef = self._get_resid_and_coef(X, xi, xj_list)
            for j, xj in enumerate(xj_list):
                B[xi, xj] = coef[j]

        # DEBAG: self._adjacency_matrix_no_lc = B.copy()

        # Check confounders
        for xi in range(n_features):
            xj_list = list(C[xi])
            xj_list.sort()
            if len(xj_list) == 0:
                continue

            for xj in xj_list:
                B[xi, xj] = np.nan

        self._adjacency_matrix = B
        return self

    def estimate_total_effect(self, X, from_index, to_index):
        # Check parameters
        X = check_array(X)

        # Check from/to ancestors
        if to_index in self._ancestors_list[from_index]:
            warnings.warn(
                f"The estimated causal effect may be incorrect because "
                f"the causal order of the destination variable (to_index={to_index}) "
                f"is earlier than the source variable (from_index={from_index})."
            )

        # Check confounders
        if True in np.isnan(self._adjacency_matrix[from_index]):
            warnings.warn(
                f"The estimated causal effect may be incorrect because "
                f"the source variable (from_index={from_index}) is influenced by confounders."
            )
            return np.nan

        # from_index + parents indices
        parents = np.where(np.abs(self._adjacency_matrix[from_index]) > 0)[0]
        predictors = [from_index]
        predictors.extend(parents)

        # Estimate total effect
        lr = LinearRegression()
        lr.fit(X[:, predictors], X[:, to_index])

        return lr.coef_[0]

    def estimate_total_effect2(self, from_index, to_index):
        # Check from/to ancestors
        if to_index in self._ancestors_list[from_index]:
            warnings.warn(
                f"The estimated causal effect may be incorrect because "
                f"the causal order of the destination variable (to_index={to_index}) "
                f"is earlier than the source variable (from_index={from_index})."
            )

        # Check confounders
        if True in np.isnan(self._adjacency_matrix[from_index]):
            warnings.warn(
                f"The estimated causal effect may be incorrect because "
                f"the source variable (from_index={from_index}) is influenced by confounders."
            )
            return np.nan

        effect = calculate_total_effect(self._adjacency_matrix, from_index, to_index)

        return effect

    def get_error_independence_p_values(self, X):
        """Calculate the p-value matrix of independence between error variables.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Original data, where n_samples is the number of samples
            and n_features is the number of features.

        Returns
        -------
        independence_p_values : array-like, shape (n_features, n_features)
            p-value matrix of independence between error variables.
        """
        # Check parameters
        X = check_array(X)
        n_samples = X.shape[0]
        n_features = X.shape[1]

        E = X - np.dot(self._adjacency_matrix, X.T).T
        nan_cols = list(set(np.argwhere(np.isnan(self._adjacency_matrix)).ravel()))
        p_values = np.zeros([n_features, n_features])
        for i, j in itertools.combinations(range(n_features), 2):
            if i in nan_cols or j in nan_cols:
                p_values[i, j] = np.nan
                p_values[j, i] = np.nan
            else:
                _, p_value = hsic_test_gamma(
                    np.reshape(E[:, i], [n_samples, 1]),
                    np.reshape(E[:, j], [n_samples, 1]),
                )
                p_values[i, j] = p_value
                p_values[j, i] = p_value

        return p_values

    @property
    def ancestors_list_(self):
        """Estimated ancestors list.

        Returns
        -------
        ancestors_list_ : array-like, shape (n_features)
            The list of causal ancestors sets, where
            n_features is the number of features.
        """
        return self._ancestors_list

    @property
    def adjacency_matrix_(self):
        """Estimated adjacency matrix.

        Returns
        -------
        adjacency_matrix_ : array-like, shape (n_features, n_features)
            The adjacency matrix B of fitted model, where
            n_features is the number of features.
            Set np.nan if order is unknown.
        """
        return self._adjacency_matrix

    def bootstrap(self, X, n_sampling):
        """Evaluate the statistical reliability of DAG based on the bootstrapping.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.
        n_sampling : int
            Number of bootstrapping samples.

        Returns
        -------
        result : BootstrapResult
            Returns the result of bootstrapping.
        """
        # Check parameters
        X = check_array(X)

        if isinstance(n_sampling, (numbers.Integral, np.integer)):
            if not 0 < n_sampling:
                raise ValueError("n_sampling must be an integer greater than 0.")
        else:
            raise ValueError("n_sampling must be an integer greater than 0.")

        # Bootstrapping
        adjacency_matrices = np.zeros([n_sampling, X.shape[1], X.shape[1]])
        total_effects = np.zeros([n_sampling, X.shape[1], X.shape[1]])
        for i in range(n_sampling):
            resampled_X = resample(X)
            self.fit(resampled_X)
            adjacency_matrices[i] = self._adjacency_matrix

            # Calculate total effects
            for to, ancestors in enumerate(self._ancestors_list):
                for from_ in ancestors:
                    total_effects[i, to, from_] = self.estimate_total_effect2(from_, to)

        return BootstrapResult(adjacency_matrices, total_effects)
